"""
Mixin classes for classes which offer previews of input and output at the _class_ level.

The intent is for mixing with :class:`pyiron_workflow.node.Node`, and for the inputs
and outputs to be IO channels there, but in principle this should function just fine
independently.

These previews need to be available at the class level so that suggestion menus and
ontologies can know how mixin classes relate to the rest of the world via input and
output without first having to instantiate them.
"""

from __future__ import annotations

import inspect
from abc import ABC, abstractmethod
from collections.abc import Callable
from functools import lru_cache
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    get_args,
)

from pyiron_snippets.dotdict import DotDict

from pyiron_workflow.data import NOT_DATA
from pyiron_workflow.logging import logger
from pyiron_workflow.output_parser import ParseOutput

if TYPE_CHECKING:
    pass


class HasIOPreview(ABC):
    """
    An interface mixin guaranteeing the class-level availability of input and output
    previews.

    E.g. for :class:`pyiron_workflow.node.Node` that have input and output channels.
    """

    @classmethod
    @abstractmethod
    def _build_inputs_preview(cls) -> dict[str, tuple[Any, Any]]:
        pass

    @classmethod
    @abstractmethod
    def _build_outputs_preview(cls) -> dict[str, Any]:
        pass

    @classmethod
    @lru_cache(maxsize=1)
    def preview_inputs(cls) -> dict[str, tuple[Any, Any]]:
        """
        Gives a class-level peek at the expected inputs.

        Returns:
            dict[str, tuple[Any, Any]]: The input name and a tuple of its
                corresponding type hint and default value.
        """
        return cls._build_inputs_preview()

    @classmethod
    @lru_cache(maxsize=1)
    def preview_outputs(cls) -> dict[str, Any]:
        """
        Gives a class-level peek at the expected outputs.

        Returns:
            dict[str, tuple[Any, Any]]: The output name and its corresponding type hint.
        """
        return cls._build_outputs_preview()

    @classmethod
    def preview_io(cls) -> DotDict[str, dict[str, Any | tuple[Any, Any]]]:
        return DotDict(
            {"inputs": cls.preview_inputs(), "outputs": cls.preview_outputs()}
        )


class ScrapesIO(HasIOPreview, ABC):
    """
    A mixin class for scraping IO channel information from a specific class method's
    signature and returns.

    Requires that the (static and class) method :meth:`_io_defining_function` be
    specified in child classes, as well as :meth:`_io_defining_function_uses_self`.
    Optionally, :attr:`_output_labels` can be overridden at the class level to avoid
    scraping the return signature for channel labels altogether.

    Since scraping returns is only possible when the function source code is available,
    this can be bypassed by manually specifying the class attribute
    :attr:`_output_labels`.

    Attributes:
        _output_labels (tuple[str] | None): The output labels to use. If None, scrapes
            them from the io-defining function. (Default is None.)
        _validate_output_labels (bool): Whether to
        _io_defining_function_uses_self (bool): Whether the signature of the IO
            defining function starts with self. When true, the first argument in the
            :meth:`_io_defining_function` is ignored. (Default is False, use the entire
            signature for specifying input.)

    Warning:
        There are a number of class features which, for computational efficiency, get
        calculated at first call and any subsequent calls return that initial value
        (including on other instances, since these are class properties); these
        depend on the :meth:`_io_defining_function` and its signature, which should
        thus be left static from the time of class definition onwards.
    """

    _extra_type_hint_scope: ClassVar[dict[str, type] | None] = None

    @classmethod
    @abstractmethod
    def _io_defining_function(cls) -> Callable:
        """Must return a static method."""

    _output_labels: ClassVar[tuple[str] | None] = None  # None: scrape them
    _validate_output_labels: ClassVar[bool] = True  # True: validate against source code
    _io_defining_function_uses_self: ClassVar[bool] = (
        False  # False: use entire signature
    )

    @classmethod
    def _build_inputs_preview(cls) -> dict[str, tuple[Any, Any]]:
        scraped: dict[str, tuple[Any, Any]] = {}
        for i, (label, value) in enumerate(cls._get_input_args().items()):
            if cls._io_defining_function_uses_self and i == 0:
                continue  # Skip the macro argument itself, it's like `self` here
            elif label in cls._get_init_keywords():
                # We allow users to parse arbitrary kwargs as channel initialization
                # So don't let them choose bad channel names
                raise ValueError(
                    f"Trying to build input preview for {cls.__name__}, encountered an "
                    f"argument name that conflicts with __init__: {label}. Please "
                    f"choose a name _not_ among {cls._get_init_keywords()}"
                )

            if value.annotation is inspect.Parameter.empty:
                type_hint = None
            elif value.annotation is None:
                type_hint = type(None)
            else:
                type_hint = value.annotation

            default = (
                NOT_DATA if value.default is inspect.Parameter.empty else value.default
            )

            scraped[label] = (type_hint, default)
        return scraped

    @classmethod
    def _build_outputs_preview(cls) -> dict[str, Any]:
        if cls._validate_output_labels:
            cls._validate()  # Validate output on first call

        labels = cls._get_output_labels()
        if labels is None:
            labels = []
        type_hints = cls._get_function_signature().return_annotation
        if type_hints is not inspect.Signature.empty:
            if type_hints is None:
                type_hints = type(None)
            if len(labels) > 1:
                type_hints = get_args(type_hints)
                if not isinstance(type_hints, tuple):
                    raise TypeError(
                        f"With multiple return labels expected to get a tuple of type "
                        f"hints, but {cls.__name__} got type {type(type_hints)}"
                    )
                if len(type_hints) != len(labels):
                    raise ValueError(
                        f"Expected type hints and return labels to have matching "
                        f"lengths, but {cls.__name__} got {len(type_hints)} hints and "
                        f"{len(labels)} labels: {type_hints}, {labels}"
                    )
            else:
                # If there's only one hint, wrap it in a tuple, so we can zip it with
                # *return_labels and iterate over both at once
                type_hints = (type_hints,)
        else:  # If there are no return hints
            type_hints = [None] * len(labels)
            # Note that this nicely differs from `NoneType`, which is the hint when
            # `None` is actually the hint!
        return dict(zip(labels, type_hints, strict=False))

    @classmethod
    def _get_output_labels(cls):
        """
        Return output labels provided for the class, scraping them from the io-defining
        function if they are not already available.
        """
        if cls._output_labels is None:
            cls._output_labels = cls._scrape_output_labels()
        return cls._output_labels

    @classmethod
    @lru_cache(maxsize=1)
    def _get_function_signature(cls) -> inspect.Signature:
        """
        The result of :func:`inspect.signature` on the io-defining function
        """
        return inspect.signature(
            cls._io_defining_function(),
            eval_str=True,
            locals=cls._extra_type_hint_scope,
        )

    @classmethod
    @lru_cache(maxsize=1)
    def _get_input_args(cls):
        return cls._get_function_signature().parameters

    @classmethod
    @lru_cache(maxsize=1)
    def _get_init_keywords(cls):
        return list(inspect.signature(cls.__init__).parameters.keys())

    @classmethod
    @lru_cache(maxsize=1)
    def _scrape_output_labels(cls):
        """
        Inspect :meth:`node_function` to scrape out strings representing the
        returned values.

         _Only_ works for functions with a single `return` expression in their body.

        It will return expressions and function calls just fine, thus good practice is
        to create well-named variables and return those so that the output labels stay
        dot-accessible.
        """
        return ParseOutput(cls._io_defining_function()).output

    @classmethod
    def _validate(cls):
        """
        Ensure that output_labels, if provided, are commensurate with graph creator
        return values, if provided, and return them as a tuple.
        """
        try:
            cls._validate_degeneracy()
            cls._validate_return_count()
        except OSError:
            logger.warn(no_output_validation_warning(cls))

    @classmethod
    def _validate_degeneracy(cls):
        output_labels = cls._get_output_labels()
        if output_labels is not None and len(set(output_labels)) != len(output_labels):
            raise ValueError(
                f"{cls.__name__} must not have degenerate output labels: "
                f"{output_labels}"
            )

    @classmethod
    def _validate_return_count(cls):
        output_labels = cls._get_output_labels()
        graph_creator_returns = ParseOutput(cls._io_defining_function()).output
        if graph_creator_returns is not None or output_labels is not None:
            error_suffix = (
                f"but {cls.__name__} got return values: {graph_creator_returns} and "
                f"labels: {output_labels}. If this intentional, you can bypass output "
                f"validation making sure the class attribute `_validate_output_labels` "
                f"is False."
            )
            try:
                if len(output_labels) != len(graph_creator_returns):
                    raise ValueError(
                        "The number of return values must exactly match the number of "
                        "output labels provided, " + error_suffix
                    )
            except TypeError as type_error:
                raise TypeError(
                    "Output labels and return values must either both or neither be "
                    "present, " + error_suffix
                ) from type_error

    @staticmethod
    def _io_defining_documentation(io_defining_function: Callable, title: str):
        """
        A helper method for building a docstring for classes that have their IO defined
        by some function.
        """
        try:
            signature = str(inspect.signature(io_defining_function))
        except Exception as e:
            signature = f"SIGNATURE NOT AVAILABLE -- {type(e).__name__}: {e}"

        try:
            source = inspect.getsource(io_defining_function)
        except Exception as e:
            source = f"SOURCE NOT AVAILABLE -- {type(e).__name__}: {e}"

        doc = (
            "" if io_defining_function.__doc__ is None else io_defining_function.__doc__
        )

        docs = f"{title.upper()} INFO:\n\n"
        docs += "Signature:\n\n"
        docs += signature
        docs += "\n\n"
        docs += "Docstring:\n\n"
        docs += doc
        docs += "\n"
        docs += "Source:\n\n"
        docs += source
        docs += "\n"
        return docs


def no_output_validation_warning(cls: type):
    return (
        f"Could not find the source code to validate {cls.__name__} output labels "
        f"against the number of returned values -- proceeding without validation"
    )



"""
:mod:`pyiron_workflow` is a python framework for constructing computational workflows
in a graph-based format.
The intent of such a framework is to improve the reliability and shareability of
computational workflows, as well as providing supporting infrastructure for the
storage and retrieval of data, and executing computations on remote resources (with a
special emphasis on HPC environments common in academic research).
It is a key goal that writing such workflows should be as easy as possible, and simple
cases should be _almost_ as simple as writing and running plain python functions.

Key features:

- Single point of import
- Easy "nodeification" of regular python code
- Macro nodes, so complex workflows can be built by composition
- (Optional) type checking for data connections
- (Optional) remote execution of individual nodes
- Both acyclic (execution flow is automated) and cyclic (execution flow must be
    specified) graphs allowed
- Easy extensibility by collecting nodes together in a python module for sharing/reusing

Planned:
- Integration with :mod:`semantikon` for ontological hinting of data channels to provide
    guided workflow design
"""

from ._version import get_versions

__version__ = get_versions()["version"]

# User API

from pyiron_workflow.api import (
    NodeSlurmExecutor,
    Workflow,  # pyironic user single-point of entry
    as_dataclass_node,
    as_function_node,
    as_macro_node,
    dataclass_node,
    for_node,
    function_node,
    macro_node,
    std,
    to_function_node,
    while_node,
)



import abc
from typing import Any, ClassVar, TypeAlias, TypeGuard, cast

from pyiron_snippets import factory

from pyiron_workflow.mixin.run import InterpretableAsExecutor
from pyiron_workflow.nodes.composite import Composite
from pyiron_workflow.nodes.static_io import StaticNode
from pyiron_workflow.storage import BackendIdentifier, StorageInterface

label_connection: TypeAlias = tuple[str, str]
label_connections: TypeAlias = tuple[label_connection, ...]
label_connections_like: TypeAlias = (
    label_connection | list[label_connection] | label_connections
)


class InvalidTestOutputError(ValueError):
    """When the test type doesn't have the right output type."""


class InvalidEdgeError(ValueError):
    """When a requested edge does not exist for the provided node types."""


class NonTerminatingLoopError(ValueError):
    """When it is strictly impossible for the loop to terminate due to a lack of recursion"""


class NotConnectionslikeError(TypeError):
    """A type checking problem."""


def _is_label_connection(c: object) -> TypeGuard[tuple[str, str]]:
    return (
        isinstance(c, tuple)
        and len(c) == 2
        and isinstance(c[0], str)
        and isinstance(c[1], str)
    )


def _tuplefy(connections: label_connections_like) -> label_connections:
    if _is_label_connection(connections):
        return (connections,)
    elif isinstance(connections, tuple) and all(
        _is_label_connection(c) for c in connections
    ):
        return cast(label_connections, connections)
    elif isinstance(connections, list) and all(
        _is_label_connection(c) for c in connections
    ):
        return tuple(connections)
    raise NotConnectionslikeError(
        f"Expected connections in the form {label_connections_like}, "
        f"but got {connections}."
    )


class While(Composite, StaticNode, abc.ABC):
    """
    A second-order, stateful, dynamic node that resets its graph at each (non-cached)
    run to dynamically add copies of a body and test case node for as long as the test
    case node is outputting a true result.

    Subclasses must define the types of nodes to use for the test and body, as well
    as at least one edge from the body to the test and from the body back to itself.

    Instances can pass executor information to all child nodes of the same type using
    the :attr:`executor_for_test` and :attr:`executor_for_body` attributes.
    """

    _body_node_class: ClassVar[type[StaticNode]]
    _test_node_class: ClassVar[type[StaticNode]]
    _body_to_test_connections: ClassVar[label_connections]
    _body_to_body_connections: ClassVar[label_connections]
    _test_stem: ClassVar[str] = "test_"
    _body_stem: ClassVar[str] = "body_"

    @classmethod
    def _build_inputs_preview(cls) -> dict[str, tuple[Any, Any]]:
        preview = {}
        for label, (hint, default) in cls._test_node_class.preview_inputs().items():
            preview[cls._test_stem + label] = (hint, default)
        for label, (hint, default) in cls._body_node_class.preview_inputs().items():
            preview[cls._body_stem + label] = (hint, default)
        preview["max_iterations"] = (int | None, None)
        return preview

    @classmethod
    def _build_outputs_preview(cls) -> dict[str, Any]:
        return dict(cls._body_node_class.preview_outputs())

    def __init__(
        self,
        *args,
        label: str | None = None,
        parent: Composite | None = None,
        autoload: BackendIdentifier | StorageInterface | None = None,
        delete_existing_savefiles: bool = False,
        autorun: bool = False,
        checkpoint: BackendIdentifier | StorageInterface | None = None,
        strict_naming: bool = True,
        executor_for_test: InterpretableAsExecutor | None = None,
        executor_for_body: InterpretableAsExecutor | None = None,
        **kwargs,
    ):
        super().__init__(
            *args,
            label=label,
            parent=parent,
            delete_existing_savefiles=delete_existing_savefiles,
            autorun=autorun,
            autoload=autoload,
            checkpoint=checkpoint,
            strict_naming=strict_naming,
            **kwargs,
        )
        self.executor_for_test = executor_for_test
        self.executor_for_body = executor_for_body

    def _on_cache_miss(self) -> None:
        super()._on_cache_miss()
        if self.ready:
            self._clean_existing_subgraph()

    def _clean_existing_subgraph(self):
        for label in self.child_labels:
            self.remove_child(label)

    def _on_run(self):
        n = 0
        test, body = self._extend_children(n)
        last_body = body  # In case of early termination, we need to link the output
        super()._on_run()
        while self._test_condition(test) and (
            self.inputs.max_iterations.value is None
            or n < self.inputs.max_iterations.value
        ):
            test >> body  # For posterity -- we manage the execution manually here
            self.starting_nodes = [body]
            super()._on_run()
            last_body = body
            n += 1
            test, body = self._extend_children(n)
            last_body >> test
            self._connect_cycles(last_body, test, body)
            super()._on_run()

        self._link_output_values(last_body)

        # Adding and removing children resets the cache, so make sure we cache _after_
        # we're done modifying the child graph
        self._cache_inputs()

        return self

    def _extend_children(self, n: int):
        test = self._test_node_class(label=f"{self._test_stem}{n}", parent=self)
        test.executor = self.executor_for_test
        body = self._body_node_class(label=f"{self._body_stem}{n}", parent=self)
        body.executor = self.executor_for_body
        self._link_input_values(test, body)
        self.starting_nodes = [test]
        return test, body

    def _link_input_values(self, test, body):
        for label, macro_input in self.inputs.items():
            if label.startswith(self._test_stem):
                macro_input.value_receiver = test.inputs[label[len(self._test_stem) :]]
            elif label.startswith(self._body_stem):
                macro_input.value_receiver = body.inputs[label[len(self._body_stem) :]]

    def _link_output_values(self, body):
        for label, body_output in body.outputs.items():
            body_output.value_receiver = self.outputs[label]

    def _connect_cycles(self, last_body, test, body):
        for target, connections in (
            (test, self._body_to_test_connections),
            (body, self._body_to_body_connections),
        ):
            for old_body_label, target_label in connections:
                last_body.outputs[old_body_label].connect(target.inputs[target_label])

    def _test_condition(self, test) -> bool:
        output_label = test.outputs.labels[0]
        return test.outputs.channel_dict[output_label].value


def _while_node_class_name(
    test_node_class: type[StaticNode],
    body_node_class: type[StaticNode],
    body_to_test_connections: label_connections,
    body_to_body_connections: label_connections,
) -> str:
    return (
        f"Do_{body_node_class.__name__}_While_{test_node_class.__name__}_With"
        f"_B2T_{'_'.join(con[0] + '2' + con[1] for con in body_to_test_connections)}"
        f"_B2B_{'_'.join(con[0] + '2' + con[1] for con in body_to_body_connections)}"
    )


@factory.classfactory
def while_node_factory(
    test_node_class: type[StaticNode],
    body_node_class: type[StaticNode],
    body_to_test_connections: label_connections,
    body_to_body_connections: label_connections,
    use_cache: bool = True,
    strict_condition_hint: bool = True,
    /,
) -> type[While]:
    """
    A factory method for creating a new :class:`While` node class.

    Args:
        test_node_class (type[StaticNode]): The class to use for the test condition.
            Must have a single output channel, and this must be hinted as boolean when
            :arg:`strict_condition_hint = True`.
        body_node_class (type[StaticNode]): The class to use for the body of the loop.
        body_to_test_connections (label_connections): Stringy representations of edge(s) from the body to the test instances.
        body_to_body_connections (label_connections): Stringy representations of edge(s) from the last-body to the next-body instances.
        use_cache (bool): Default class behavior for whether to skip running this node when its inputs match cached values. (Default is True, try to exploit caching.)
        strict_condition_hint (bool): Whether to demand that the :arg:`test_node_class` has its single output hinted as boolean. (Default is True.)

    Returns:
        type[While]: A new :class:`While` node class.
    """
    _verify_test_output(test_node_class, strict_condition_hint=strict_condition_hint)
    if len(body_to_test_connections) == 0:
        raise NonTerminatingLoopError(
            f"While-node received no connections from body ({body_node_class.__name__}) to test ({test_node_class.__name__}), and thus cannot transition from a non-terminating to terminating state. Please provide edges."
        )
    if len(body_to_body_connections) == 0:
        raise NonTerminatingLoopError(
            f"While-node received no connections from body ({body_node_class.__name__}) back to itself, and thus cannot transition from a non-terminating to terminating state. Please provide edges."
        )
    _verify_edges_exist(body_node_class, test_node_class, body_to_test_connections)
    _verify_edges_exist(body_node_class, body_node_class, body_to_body_connections)

    combined_docstring = (
        "While node docstring:\n"
        + (While.__doc__ if While.__doc__ is not None else "")
        + "\nBody node docstring:\n"
        + (body_node_class.__doc__ if body_node_class.__doc__ is not None else "")
        + "\nTest node docstring:\n"
        + (test_node_class.__doc__ if test_node_class.__doc__ is not None else "")
    )
    return (  # type: ignore[return-value]
        _while_node_class_name(
            test_node_class,
            body_node_class,
            body_to_test_connections,
            body_to_body_connections,
        ),
        (While,),
        {
            "_test_node_class": test_node_class,
            "_body_node_class": body_node_class,
            "_body_to_test_connections": body_to_test_connections,
            "_body_to_body_connections": body_to_body_connections,
            "__doc__": combined_docstring,
            "use_cache": use_cache,
        },
        {},
    )


def _verify_test_output(
    test_node_class: type[StaticNode], strict_condition_hint: bool = True
):
    test_outputs = test_node_class.preview_outputs()
    if len(test_outputs) != 1 or (
        strict_condition_hint and next(iter(test_outputs.values())) is not bool
    ):
        raise InvalidTestOutputError(
            f"While-loop Test node class {test_node_class.__name__} must have a single "
            f"boolean output channel when `strict_condition_hint = True`, but has "
            f"outputs {test_outputs}."
        )


def _verify_edges_exist(
    from_class: type[StaticNode], to_class: type[StaticNode], edges: label_connections
):
    for from_label, to_label in edges:
        if from_label not in from_class.preview_outputs():
            raise InvalidEdgeError(
                f"While-loop body node class {from_class.__name__} has no output "
                f"channel {from_label}."
            )
        if to_label not in to_class.preview_inputs():
            raise InvalidEdgeError(
                f"While-loop test or body node class {to_class.__name__} has no input {to_label}."
            )


def while_node(
    test_node_class: type[StaticNode],
    body_node_class: type[StaticNode],
    body_to_test_connections: label_connections_like,
    body_to_body_connections: label_connections_like,
    *node_args,
    use_cache: bool = True,
    strict_condition_hint: bool = True,
    **node_kwargs,
):
    """
    Make a new
    Args:
        test_node_class (type[StaticNode]): The class to use for the test condition.
            Must have a single output channel, and this must be hinted as boolean when
            :arg:`strict_condition_hint = True`.
        body_node_class (type[StaticNode]): The class to use for the body of the loop.
        body_to_test_connections (label_connections): Stringy representations of edge(s) from the body to the test instances.
        body_to_body_connections (label_connections): Stringy representations of edge(s) from the last-body to the next-body instances.
        *node_args: All the usual parent class args.
        use_cache (bool): Default class behavior for whether to skip running this node when its inputs match cached values. (Default is True, try to exploit caching.)
        strict_condition_hint (bool): Whether to demand that the :arg:`test_node_class` has its single output hinted as boolean. (Default is True.)
        **node_kwargs: All the usual parent class kwargs, including :arg:`executor_for_test` and :arg:`executor_for_body` that are specific to :class:`While`.

    Returns:
        While: An instance of the new :class:`While` subclass.

    Examples:
        >>> from pyiron_workflow import Workflow
        >>> wf = Workflow("my_while_loop")
        >>> wf.x0 = Workflow.create.std.UserInput(0)
        >>> wf.limit = Workflow.create.std.UserInput(5)
        >>> wf.step = Workflow.create.std.UserInput(2)
        >>> wf.add_while = Workflow.create.while_node(
        ...     Workflow.create.std.LessThan,  # Test
        ...     Workflow.create.std.Add,  # Body
        ...     (("add", "obj"),),  # body-to-test
        ...     (("add", "obj"),),  # body-to-body
        ...     strict_condition_hint=False,  # LessThan doesn't hint it's boolean return...
        ...     test_obj=wf.x0,
        ...     test_other=wf.limit,
        ...     body_obj=wf.x0,
        ...     body_other=wf.step
        ... )
        >>> wf()
        {'add_while__add': 6}

        Note how the while-node input maps the test and body node inputs with a pre-pended identifier, while the output maps the body node outputs directly.
    """
    b2t = _tuplefy(body_to_test_connections)
    b2b = _tuplefy(body_to_body_connections)
    while_node_factory.clear(
        _while_node_class_name(test_node_class, body_node_class, b2t, b2b)
    )
    cls = while_node_factory(
        test_node_class,
        body_node_class,
        b2t,
        b2b,
        use_cache,
        strict_condition_hint,
    )
    return cls(*node_args, **node_kwargs)



"""
A utility for finding public `pyiron_workflow.node.Node` objects.

Supports the idea of node developers writing independent node packages.
"""

from __future__ import annotations

import importlib.util
import inspect
import sys
from pathlib import Path
from types import ModuleType
from typing import TypeVar, cast

from pyiron_workflow.node import Node

NodeType = TypeVar("NodeType", bound=Node)


def _get_subclasses(
    source: str | Path | ModuleType,
    base_class: type[NodeType],
    get_private: bool = False,
    get_abstract: bool = False,
    get_imports_too: bool = False,
) -> list[type[NodeType]]:
    if isinstance(source, str | Path):
        source = Path(source)
        if source.is_file():
            # Load the module from the file
            module_name = source.stem
            spec = importlib.util.spec_from_file_location(module_name, str(source))
            if spec is None or spec.loader is None:
                raise ImportError(f"Could not create a ModuleSpec for {source}")
            module = importlib.util.module_from_spec(spec)
            sys.modules[module_name] = module
            spec.loader.exec_module(module)
        else:
            raise ValueError("File path does not point to a valid file")
    elif inspect.ismodule(source):
        module = source
    else:
        raise ValueError("Input must be a module or a valid file path")

    return [
        obj
        for name, obj in inspect.getmembers(module, inspect.isclass)
        if (
            issubclass(obj, base_class)
            and (get_private or not name.startswith("_"))
            and (get_abstract or not inspect.isabstract(obj))
            and (get_imports_too or _locally_defined(obj, module))
        )
    ]


def _locally_defined(obj, module):
    obj_module_name = obj.__module__
    obj_module = importlib.import_module(obj_module_name)
    return obj_module.__file__ == module.__file__


def find_nodes(source: str | Path | ModuleType) -> list[type[Node]]:
    """
    Get a list of all public, non-abstract nodes defined in the source.
    """
    return cast(list[type[Node]], _get_subclasses(source, Node))



from __future__ import annotations

import itertools
import math
from abc import ABC
from concurrent.futures import Executor
from functools import lru_cache
from typing import Any, ClassVar

from pandas import DataFrame
from pyiron_snippets.factory import classfactory

from pyiron_workflow.data import NOT_DATA
from pyiron_workflow.mixin.run import InterpretableAsExecutor
from pyiron_workflow.nodes.composite import Composite
from pyiron_workflow.nodes.static_io import StaticNode
from pyiron_workflow.nodes.transform import (
    InputsToDict,
    inputs_to_dataframe,
    inputs_to_dict,
    inputs_to_list,
)
from pyiron_workflow.storage import BackendIdentifier, StorageInterface


def dictionary_to_index_maps(
    data: dict,
    nested_keys: list[str] | tuple[str, ...] | None = None,
    zipped_keys: list[str] | tuple[str, ...] | None = None,
):
    """
    Given a dictionary where some data is iterable, and list(s) of keys over
    which to make a nested and/or zipped loop, return dictionaries mapping
    these keys to all the indices of the data they hold. Zipped loops are
    nested outside the nesting loops.

    Args:
        data (dict): The dictionary of data, some of which must me iterable.
        nested_keys (tuple[str, ...] | None): The keys whose data to make a
            nested for-loop over.
        zipped_keys (tuple[str, ...] | None): The keys whose data to make a
            zipped for-loop over.

    Returns:
        (tuple[dict[..., int], ...]): A tuple of dictionaries where each item
            maps the dictionary key to an index for that key's value.

    Raises:
        (KeyError): If any of the provided keys are not keys of the provided
            dictionary.
        (TypeError): If any of the data held in a provided key does cannot be
            operated on with `len`.
        (ValueError): If neither set of keys to iterate on is provided, or if
            all values being iterated over have a length of zero.
    """

    try:
        nested_data_lengths = (
            []
            if (nested_keys is None or len(nested_keys) == 0)
            else [len(data[key]) for key in nested_keys]
        )
    except TypeError as e:
        raise TypeError(
            f"Could not parse nested lengths -- Does one of the keys {nested_keys} "
            f"have non-iterable data?"
        ) from e
    n_nest = math.prod(nested_data_lengths) if len(nested_data_lengths) > 0 else 0

    try:
        n_zip = (
            0
            if (zipped_keys is None or len(zipped_keys) == 0)
            else min(len(data[key]) for key in zipped_keys)
        )
    except TypeError as e:
        raise TypeError(
            f"Could not parse zipped lengths -- Does one of the keys {zipped_keys} "
            f"have non-iterable data?"
        ) from e

    def nested_generator():
        return itertools.product(*[range(n) for n in nested_data_lengths])

    def nested_index_map(nested_indices):
        return {
            nested_keys[i_key]: nested_index
            for i_key, nested_index in enumerate(nested_indices)
        }

    def zipped_generator():
        return range(n_zip)

    def zipped_index_map(zipped_index):
        return dict.fromkeys(zipped_keys, zipped_index)

    def merge(d1, d2):
        d1.update(d2)
        return d1

    if n_nest > 0 and n_zip > 0:
        key_index_maps = tuple(
            merge(nested_index_map(nested_indices), zipped_index_map(zipped_index))
            for nested_indices, zipped_index in itertools.product(
                nested_generator(), zipped_generator()
            )
        )
    elif n_nest > 0:
        key_index_maps = tuple(
            nested_index_map(nested_indices) for nested_indices in nested_generator()
        )
    elif n_zip > 0:
        key_index_maps = tuple(
            zipped_index_map(zipped_index) for zipped_index in zipped_generator()
        )
    elif nested_keys is None and zipped_keys is None:
        raise ValueError(
            "At least one of `nested_keys` or `zipped_keys` must be specified."
        )
    else:
        raise ValueError("Received keys to iterate over, but all values had length 0.")

    return key_index_maps


class UnmappedConflictError(ValueError):
    """
    When a for-node gets a body whose output label conflicts with looped a input
    label and no map was provided to avoid this.
    """


class MapsToNonexistentOutputError(ValueError):
    """
    When a for-node tries to map body node output channels that don't exist.
    """


class For(Composite, StaticNode, ABC):
    """
    Specifies fixed fields of some other node class to iterate over, but allows the
    length of looped input to vary by dynamically destroying and recreating (most of)
    its subgraph at run-time.

    Collects looped output and collates them with looped input values in a dataframe.

    The :attr:`body_node_executor` gets applied to each body node instance on each
    run.
    """

    _body_node_class: ClassVar[type[StaticNode]]
    _iter_on: ClassVar[tuple[str, ...]] = ()
    _zip_on: ClassVar[tuple[str, ...]] = ()
    _output_as_dataframe: ClassVar[bool] = True
    _output_column_map: ClassVar[dict[str, str]] = {}

    def __init_subclass__(cls, output_column_map=None, **kwargs):
        super().__init_subclass__(**kwargs)

        unmapped_conflicts = (
            set(cls._body_node_class.preview_inputs().keys())
            .intersection(cls._iter_on + cls._zip_on)
            .intersection(cls._body_node_class.preview_outputs().keys())
            .difference(() if output_column_map is None else output_column_map.keys())
        )
        if len(unmapped_conflicts) > 0:
            raise UnmappedConflictError(
                f"The body node {cls._body_node_class.__name__} has channel labels "
                f"{unmapped_conflicts} that appear as both (looped) input _and_ output "
                f"for {cls.__name__}. All such channels require a map to produce new, "
                f"unique column names for the output."
            )

        maps_to_nonexistent_output = set(
            {} if output_column_map is None else output_column_map.keys()
        ).difference(cls._body_node_class.preview_outputs().keys())
        if len(maps_to_nonexistent_output) > 0:
            raise MapsToNonexistentOutputError(
                f"{cls.__name__} tried to map body node output(s) "
                f"{maps_to_nonexistent_output} to new column names, but "
                f"{cls._body_node_class.__name__} has no such outputs."
            )

        cls._output_column_map = {} if output_column_map is None else output_column_map

    @classmethod
    @lru_cache(maxsize=1)
    def output_column_map(cls) -> dict[str, str]:
        """
        How to transform body node output labels to dataframe column names.
        """
        map_ = {k: k for k in cls._body_node_class.preview_outputs()}
        for body_label, column_name in cls._output_column_map.items():
            map_[body_label] = column_name
        return map_

    def __init__(
        self,
        *args,
        label: str | None = None,
        parent: Composite | None = None,
        autoload: BackendIdentifier | StorageInterface | None = None,
        delete_existing_savefiles: bool = False,
        autorun: bool = False,
        checkpoint: BackendIdentifier | StorageInterface | None = None,
        strict_naming: bool = True,
        body_node_executor: InterpretableAsExecutor | None = None,
        **kwargs,
    ):
        super().__init__(
            *args,
            label=label,
            parent=parent,
            delete_existing_savefiles=delete_existing_savefiles,
            autorun=autorun,
            autoload=autoload,
            checkpoint=checkpoint,
            strict_naming=strict_naming,
            **kwargs,
        )
        self.body_node_executor = body_node_executor

    def _setup_node(self) -> None:
        super()._setup_node()
        input_nodes = []
        for channel in self.inputs:
            n = self.create.std.UserInput(
                channel.default, label=channel.label, parent=self
            )
            n.inputs.user_input.type_hint = channel.type_hint
            channel.value_receiver = n.inputs.user_input
            input_nodes.append(n)
        self.starting_nodes = input_nodes
        self._input_node_labels = tuple(n.label for n in input_nodes)

    def _on_cache_miss(self) -> None:
        super()._on_cache_miss()
        if self.ready:
            self._build_body()

    def _build_body(self):
        """
        Construct instances of the body node based on input length, and wire them to IO.
        """
        iter_maps = dictionary_to_index_maps(
            self.inputs.to_value_dict(),
            nested_keys=self._iter_on,
            zipped_keys=self._zip_on,
        )

        self._clean_existing_subgraph()

        self._create_and_connect_input_to_body_nodes(iter_maps)

        if self._output_as_dataframe:
            self._collect_output_as_dataframe(iter_maps)
        else:
            self._collect_output_as_lists(iter_maps)

        self.set_run_signals_to_dag_execution()

    def _clean_existing_subgraph(self):
        for label in self.child_labels:
            if label not in self._input_node_labels:
                self.remove_child(label)
            else:
                # Re-run the user input node so it has up-to-date output, otherwise
                # when we inject a getitem node -- which will try to run automatically
                # -- it will see data it can work with, but if that data happens to
                # have the wrong length it may successfully auto-run on the wrong thing
                # and throw an error!
                self.children[label].run(
                    run_data_tree=False,
                    run_parent_trees_too=False,
                    fetch_input=False,
                    emit_ran_signal=False,
                    # Data should simply be coming from the value link
                    # We just want to refresh the output
                )
        # TODO: Instead of deleting _everything_ each time, try and re-use stuff

    def _create_and_connect_input_to_body_nodes(self, iter_maps):
        for n, channel_map in enumerate(iter_maps):
            # Make the body node
            body_node = self._body_node_class(label=self._body_name(n), parent=self)
            body_node.executor = self.body_node_executor

            # Broadcast macro input to body node
            for bcast_label in set(self.preview_inputs().keys()).difference(
                self._iter_on + self._zip_on
            ):
                self.inputs[bcast_label].value_receiver = body_node.inputs[bcast_label]

            # Get item from macro input and connect it to body node
            for looped_label, i in channel_map.items():
                index_node = self.children[looped_label][i]  # Inject getitem node
                body_node.inputs[looped_label] = index_node

    @staticmethod
    def _body_name(n: int):
        return f"body_{n}"

    def _collect_output_as_dataframe(self, iter_maps):
        self.dataframe = inputs_to_dataframe(len(iter_maps))
        self.dataframe.outputs.df.value_receiver = self.outputs.df

        for n, channel_map in enumerate(iter_maps):
            row_collector = self._build_row_collector_node(n)
            for label, i in channel_map.items():
                row_collector.inputs[label] = self.children[label][i]

            for label, body_out in self[self._body_name(n)].outputs.items():
                row_collector.inputs[self.output_column_map()[label]] = body_out

            self.dataframe.inputs[f"row_{n}"] = row_collector

    def _build_row_collector_node(self, row_number) -> InputsToDict:
        # Iterated inputs
        row_specification = {
            key: (self._body_node_class.preview_inputs()[key][0], NOT_DATA)
            for key in self._iter_on + self._zip_on
        }
        # Outputs
        row_specification.update(
            {
                self.output_column_map()[key]: (hint, NOT_DATA)
                for key, hint in self._body_node_class.preview_outputs().items()
            }
        )
        return inputs_to_dict(
            row_specification, parent=self, label=f"row_collector_{row_number}"
        )

    def _collect_output_as_lists(self, iter_maps):
        n_rows = len(iter_maps)

        def column_collector_name(s: str):
            return f"column_collector_{s}"

        for label, hint in self._body_node_class.preview_outputs().items():
            mapped_label = self.output_column_map()[label]
            column_collector = inputs_to_list(
                n_rows,
                label=column_collector_name(mapped_label),
                parent=self,
            )
            column_collector.outputs.list.type_hint = (
                list if hint is None else list[hint]
            )
            column_collector.outputs.list.value_receiver = self.outputs[mapped_label]

            for n, _ in enumerate(iter_maps):
                column_collector.inputs[f"item_{n}"] = self[self._body_name(n)].outputs[
                    label
                ]

        for label in self._zip_on + self._iter_on:
            column_collector = inputs_to_list(
                n_rows,
                label=column_collector_name(label),
                parent=self,
            )
            hint = self._body_node_class.preview_inputs()[label][0]
            column_collector.outputs.list.type_hint = (
                list if hint is None else list[hint]
            )
            column_collector.outputs.list.value_receiver = self.outputs[label]
        for n, channel_map in enumerate(iter_maps):
            for label, i in channel_map.items():
                self[column_collector_name(label)].inputs[f"item_{n}"] = self[label][i]

    @classmethod
    @lru_cache(maxsize=1)
    def _build_inputs_preview(cls) -> dict[str, tuple[Any, Any]]:
        preview = {}
        for label, (hint, default) in cls._body_node_class.preview_inputs().items():
            # TODO: Leverage hint and default, listing if it's looped on
            if label in cls._zip_on + cls._iter_on:
                preview[label] = (
                    list if hint is None else list[hint],  # type: ignore[valid-type]
                    NOT_DATA,  # TODO: Figure out a generator pattern to get lists
                )
            else:
                preview[label] = (hint, default)
        return preview

    @classmethod
    def _build_outputs_preview(cls) -> dict[str, Any]:
        if cls._output_as_dataframe:
            return {"df": DataFrame}
        else:
            preview = {}
            for label, (
                hint,
                _default,
            ) in cls._body_node_class.preview_inputs().items():
                if label in cls._zip_on + cls._iter_on:
                    preview[label] = list if hint is None else list[hint]  # type: ignore[valid-type]
            for label, hint in cls._body_node_class.preview_outputs().items():
                preview[cls.output_column_map()[label]] = (
                    list if hint is None else list[hint]  # type: ignore[valid-type]
                )
            return preview

    @property
    def _input_value_links(self):
        """
        Value connections between child output and macro in string representation based
        on labels.

        The string representation helps storage, and having it as a property ensures
        the name is protected.
        """
        # Duplicated value linking behaviour from Macro class
        return [
            (c.label, (c.value_receiver.owner.label, c.value_receiver.label))
            for c in self.inputs
        ]

    @property
    def _output_value_links(self):
        """
        Value connections between macro and child input in string representation based
        on labels.

        The string representation helps storage, and having it as a property ensures
        the name is protected.
        """
        # Duplicated value linking behaviour from Macro class
        return [
            ((c.owner.label, c.label), c.value_receiver.label)
            for child in self
            for c in child.outputs
            if c.value_receiver is not None
        ]

    def __getstate__(self):
        state = super().__getstate__()
        if isinstance(self.body_node_executor, Executor):
            state["body_node_executor"] = None

        # Duplicated value linking behaviour from Macro class
        state["_input_value_links"] = self._input_value_links
        state["_output_value_links"] = self._output_value_links

        return state

    def _get_state_from_remote_other(self, other_self):
        state = super()._get_state_from_remote_other(other_self)
        if state["body_node_executor"] is None and self.body_node_executor is not None:
            state.pop("body_node_executor")  # May have been overridden to None for
            # __getstate__ to avoid serializing an executor instance, so keep local
        return state

    def __setstate__(self, state):
        # Duplicated value linking behaviour from Macro class
        # Purge value links from the state
        input_links = state.pop("_input_value_links")
        output_links = state.pop("_output_value_links")

        super().__setstate__(state)

        # Re-forge value links
        for inp, (child, child_inp) in input_links:
            self.inputs[inp].value_receiver = self.children[child].inputs[child_inp]

        for (child, child_out), out in output_links:
            self.children[child].outputs[child_out].value_receiver = self.outputs[out]

    @property
    def nrows(self) -> int | None:
        if self.outputs.ready:
            if self._output_as_dataframe:
                return len(self.outputs.df.value)
            else:
                return len(self.outputs[self.outputs.labels[0]].value)
        else:
            return None

    @property
    def ncols(self) -> int | None:
        if self.outputs.ready:
            if self._output_as_dataframe:
                return len(self.outputs.df.value.columns)
            else:
                return len(self.outputs)
        else:
            return None


def _for_node_class_name(
    body_node_class: type[StaticNode],
    iter_on: tuple[str, ...],
    zip_on: tuple[str, ...],
    output_as_dataframe: bool,
):
    iter_fields = (
        "" if len(iter_on) == 0 else "Iter" + "".join(k.title() for k in iter_on)
    )
    zip_fields = "" if len(zip_on) == 0 else "Zip" + "".join(k.title() for k in zip_on)
    out = "DataOut" if output_as_dataframe else "ListOut"
    return f"{For.__name__}{body_node_class.__name__}{iter_fields}{zip_fields}{out}"


@classfactory
def for_node_factory(
    body_node_class: type[StaticNode],
    iter_on: tuple[str, ...] | str = (),
    zip_on: tuple[str, ...] | str = (),
    output_as_dataframe: bool = True,
    output_column_map: dict | None = None,
    use_cache: bool = True,
    /,
) -> type[For]:
    combined_docstring = (
        "For node docstring:\n"
        + (For.__doc__ if For.__doc__ is not None else "")
        + "\nBody node docstring:\n"
        + (body_node_class.__doc__ if body_node_class.__doc__ is not None else "")
    )

    iter_on = (iter_on,) if isinstance(iter_on, str) else iter_on
    zip_on = (zip_on,) if isinstance(zip_on, str) else zip_on

    return (  # type: ignore[return-value]
        _for_node_class_name(body_node_class, iter_on, zip_on, output_as_dataframe),
        (For,),
        {
            "_body_node_class": body_node_class,
            "_iter_on": iter_on,
            "_zip_on": zip_on,
            "_output_as_dataframe": output_as_dataframe,
            "__doc__": combined_docstring,
            "use_cache": use_cache,
        },
        {"output_column_map": output_column_map},
    )


def for_node(
    body_node_class: type[StaticNode],
    *node_args,
    iter_on: tuple[str, ...] | str = (),
    zip_on: tuple[str, ...] | str = (),
    output_as_dataframe: bool = True,
    output_column_map: dict[str, str] | None = None,
    use_cache: bool = True,
    **node_kwargs,
):
    """
    Makes a new :class:`For` node which internally creates instances of the
    :param:`body_node_class` and loops input onto them in nested and/or zipped loop(s).

    Output is a single channel, `"df"`, which holds a :class:`pandas.DataFrame` whose
    rows couple (looped) input to their respective body node outputs.

    The internal node structure gets re-created each run, so the same inputs must
    consistently be iterated over, but their lengths can change freely.

    An executor can be applied to all body node instances at run-time by assigning it
    to the :attr:`body_node_executor` attribute of the for-node.

    Args:
        body_node_class type[StaticNode]: The class of node to loop on.
        *node_args: Regular positional node arguments.
        iter_on (tuple[str, ...] | str): Input label(s) in the :param:`body_node_class`
            to nested-loop on.
        zip_on (tuple[str, ...] | str): Input label(s) in the :param:`body_node_class`
            to zip-loop on.
        output_as_dataframe (bool): Whether to package the output (and iterated input)
            as a dataframe, or leave them as individual lists. (Default is True,
            package as dataframe.)
        output_column_map (dict[str, str] | None): A map for generating dataframe
            column names (values) from body node output channel labels (keys).
            Necessary iff the body node has the same label for an output channel and
            an input channel being looped over. (Default is None, just use the output
            channel labels as columb names.)
        use_cache (bool): Whether this node should default to caching its values.
            (Default is True.)
        **node_kwargs: Regular keyword node arguments.

    Returns:
        (For): An instance of a dynamically-subclassed :class:`For` node.

    Examples:
        >>> from pyiron_workflow import Workflow
        >>>
        >>> @Workflow.wrap.as_function_node("together")
        ... def FiveTogether(a: int, b: int, c: int, d: int, e: str = "foobar"):
        ...     return (a, b, c, d, e),
        >>>
        >>> for_instance = Workflow.create.for_node(
        ...     FiveTogether,
        ...     iter_on=("a", "b"),
        ...     zip_on=("c", "d"),
        ...     a=[1, 2],
        ...     b=[3, 4, 5, 6],
        ...     c=[7, 8],
        ...     d=[9, 10, 11],
        ...     e="e"
        ... )
        >>>
        >>> out = for_instance()
        >>> type(out.df)
        <class 'pandas.core.frame.DataFrame'>

        Internally, the loop node has made a bunch of body nodes, as well as nodes to
        index and collect data
        >>> len(for_instance)
        48

        We get one dataframe row for each possible combination of looped input
        >>> len(out.df)
        16

        We are stuck iterating on the fields we defined, but we can change the length
        of the input and the loop node's body will get reconstructed at run-time to
        accommodate this
        >>> out = for_instance(a=[1], b=[3], d=[7])
        >>> len(for_instance), len(out)
        (12, 1)

        Note that if we had simply returned each input individually, without any output
        labels on the node, we'd need to specify a map on the for-node so that the
        (looped) input and output columns on the resulting dataframe are all unique:
        >>> @Workflow.wrap.as_function_node
        ... def FiveApart(a: int, b: int, c: int, d: int, e: str = "foobar"):
        ...     return a, b, c, d, e,
        >>>
        >>> for_instance = Workflow.create.for_node(
        ...     FiveApart,
        ...     iter_on=("a", "b"),
        ...     zip_on=("c", "d"),
        ...     a=[1, 2],
        ...     b=[3, 4, 5, 6],
        ...     c=[7, 8],
        ...     d=[9, 10, 11],
        ...     e="e",
        ...     output_column_map={
        ...         "a": "out_a",
        ...         "b": "out_b",
        ...         "c": "out_c",
        ...         "d": "out_d"
        ...     }
        ... )
        >>>
        >>> out = for_instance()
        >>> out.df.columns
        Index(['a', 'b', 'c', 'd', 'out_a', 'out_b', 'out_c', 'out_d', 'e'], dtype='object')

    """
    iter_on = (iter_on,) if isinstance(iter_on, str) else iter_on
    zip_on = (zip_on,) if isinstance(zip_on, str) else zip_on
    for_node_factory.clear(
        _for_node_class_name(body_node_class, iter_on, zip_on, output_as_dataframe)
    )
    cls = for_node_factory(
        body_node_class,
        iter_on,
        zip_on,
        output_as_dataframe,
        output_column_map,
        use_cache,
    )
    cls.preview_io()
    return cls(*node_args, **node_kwargs)



"""
Simple display capabilities to make it easier for humans to see what's happening.
"""

from abc import ABC
from json import dumps
from typing import Any

from pyiron_workflow.mixin.has_interface_mixins import UsesState


class HasStateDisplay(UsesState, ABC):
    """
    A mixin that leverages :meth:`__getstate__` to automatically build a half-decent
    JSON-compatible representation dictionary.

    Child classes can over-ride :meth:`display_state` to add or remove elements from
    the display dictionary, e.g. to (optionally) expose state elements that would
    otherwise be private or to show properties that are computed and not stored in
    state, or (mandatory -- JSON demands it) remove recursion from the state.

    Provides a :meth:`_repr_json_` method leveraging this beautified state dictionary
    to give a standard JSON representation in Jupyter notebooks.
    """

    def display_state(
        self, state: dict | None = None, ignore_private: bool = True
    ) -> dict[str, Any]:
        """
        A dictionary of JSON-compatible objects based on the object state (plus
        whatever modifications to the state the class designer has chosen to make).

        Anything that fails to dump to JSON gets cast as a string and then dumped.

        Args:
            state (dict|None): The starting state. Default is None which uses
                `__getstate__`, but available in case child classes want to first
                sanitize the state values.
            ignore_private (bool): Whether to ignore or include any state element
                whose key starts with `'_'`. Default is True, only show public data.

        Returns:
            dict:
        """
        display = dict(self.__getstate__()) if state is None else state
        to_del = []
        for k, v in display.items():
            if ignore_private and k.startswith("_"):
                to_del.append(k)
                continue

            if isinstance(v, HasStateDisplay):
                display[k] = v.display_state(ignore_private=ignore_private)
            else:
                try:
                    display[k] = dumps(v)
                except TypeError:
                    display[k] = dumps(str(v))

        for k in to_del:
            del display[k]

        return display

    def _repr_json_(self):
        return self.display_state()

    @staticmethod
    def _make_entry_public(state: dict, private_key: str, public_key: str):
        if private_key not in state:
            raise ValueError(
                f"Can't make {private_key} public, it was not found among "
                f"{list(state.keys())}"
            )
        if public_key in state:
            raise ValueError(
                f"Can't make {private_key} public, {public_key} is already a key in"
                f" the dict!"
            )
        state[public_key] = state[private_key]
        del state[private_key]
        return state



from __future__ import annotations

"""
Shared code for various node-creating decorators facilitating multiple dispatch (i.e.
using decorators with or without arguments, contextually.
"""


class MultipleDispatchError(ValueError):
    """
    Raise from callables using multiple dispatch when no interpretation of input
    matches an expected case.
    """


def dispatch_output_labels(single_dispatch_decorator):
    def multi_dispatch_decorator(*output_labels, **kwargs):
        if len(output_labels) > 0 and callable(output_labels[0]):
            if len(output_labels) > 1:
                raise MultipleDispatchError(
                    f"Output labels must all be strings (for decorator usage with an "
                    f"argument), or a callable must be provided alone -- got "
                    f"{output_labels}."
                )
            return single_dispatch_decorator(**kwargs)(output_labels[0])
        else:
            return single_dispatch_decorator(*output_labels, **kwargs)

    return multi_dispatch_decorator



"""
This module holds customized children of :class:`concurrent.futures.Executor`.
"""

from pyiron_workflow.executors.cloudpickleprocesspool import (
    CloudpickleProcessPoolExecutor,
)
from pyiron_workflow.executors.wrapped_executorlib import NodeSlurmExecutor



import inspect
from typing import Any, ClassVar

from executorlib import BaseExecutor, SingleNodeExecutor, SlurmClusterExecutor
from executorlib.api import TestClusterExecutor

from pyiron_workflow.mixin import lexical, run


class DedicatedExecutorError(TypeError):
    """
    To raise when you try to use one of these executors outside the context of a node.
    """


class ProtectedResourceError(ValueError):
    """
    Raise when a user provides executorlib resources that we need to override.
    """


class CacheOverride(BaseExecutor):
    override_cache_file_name: ClassVar[str] = "executorlib_cache"

    def submit(self, fn, /, *args, **kwargs):
        """
        We demand that `fn` be the bound-method `on_run` of a `Lexical`+`Runnable`
        class (a `Node` is, of course, the intended resolution of this demand).
        """
        if (
            inspect.ismethod(fn)
            and fn.__name__ == "on_run"
            and isinstance(fn.__self__, lexical.Lexical)  # provides .as_path
            and isinstance(fn.__self__, run.Runnable)  # provides .on_run
        ):
            cache_key_info = {
                "cache_key": self.override_cache_file_name,
                "cache_directory": str(fn.__self__.as_path()),
            }
        else:
            raise DedicatedExecutorError(
                f"{self.__class__.__name__} is only intended to work with the "
                f"on_run method of pyiron_workflow.Node objects, but got {fn}"
            )

        _validate_existing_resource_dict(kwargs)

        if "resource_dict" in kwargs:
            kwargs["resource_dict"].update(cache_key_info)
        else:
            kwargs["resource_dict"] = cache_key_info

        return super().submit(fn, *args, **kwargs)


def _validate_existing_resource_dict(kwargs: dict[str, Any]):
    if "resource_dict" in kwargs:
        if "cache_key" in kwargs["resource_dict"]:
            raise ProtectedResourceError(
                f"pyiron_workflow needs the freedom to specify the cache, so the "
                f'requested "cache_directory" '
                f"({kwargs['resource_dict']['cache_key']}) would get overwritten."
            )
        if "cache_directory" in kwargs["resource_dict"]:
            raise ProtectedResourceError(
                f"pyiron_workflow needs the freedom to specify the cache, so the "
                f'requested "cache_directory" '
                f"({kwargs['resource_dict']['cache_directory']})would get "
                f"overwritten."
            )


class NodeSingleExecutor(CacheOverride, SingleNodeExecutor): ...


class NodeSlurmExecutor(CacheOverride, SlurmClusterExecutor): ...


class _CacheTestClusterExecutor(CacheOverride, TestClusterExecutor): ...



"""
A base class for objects that can form nodes in the graph representation of a
computational workflow.

The workhorse class for the entire concept.
"""

from __future__ import annotations

import contextlib
import pathlib
import shutil
from abc import ABC, abstractmethod
from concurrent.futures import Future
from importlib import import_module
from typing import TYPE_CHECKING, Any, Literal, cast

from pyiron_snippets.colors import SeabornColors
from pyiron_snippets.dotdict import DotDict

from pyiron_workflow.channels import (
    AccumulatingInputSignal,
    Channel,
    InputLockedError,
    InputSignal,
    OutputSignal,
)
from pyiron_workflow.draw import Node as GraphvizNode
from pyiron_workflow.executors.wrapped_executorlib import CacheOverride
from pyiron_workflow.io import IO, Inputs, Signals
from pyiron_workflow.logging import logger
from pyiron_workflow.mixin.display_state import HasStateDisplay
from pyiron_workflow.mixin.injection import (
    InjectsOnChannel,
    OutputDataWithInjection,
    OutputsWithInjection,
)
from pyiron_workflow.mixin.lexical import Lexical
from pyiron_workflow.mixin.run import ReadinessError, Runnable
from pyiron_workflow.storage import (
    BackendIdentifier,
    StorageInterface,
    available_backends,
)
from pyiron_workflow.topology import (
    get_nodes_in_data_tree,
    set_run_connections_according_to_linear_dag,
)

if TYPE_CHECKING:
    from pathlib import Path

    import graphviz

    from pyiron_workflow.nodes.composite import Composite


class WaitingForFutureError(ValueError): ...


class AmbiguousOutputError(ValueError):
    """Raised when searching for exactly one output, but multiple are found."""


class ConnectionCopyError(ValueError):
    """Raised when trying to copy IO, but connections cannot be copied"""


class ValueCopyError(ValueError):
    """Raised when trying to copy IO, but values cannot be copied"""


class Node(
    HasStateDisplay,
    Lexical["Composite"],
    Runnable,
    InjectsOnChannel,
    ABC,
):
    """
    Nodes are elements of a computational graph.
    They have inputs and outputs to interface with the wider world, and perform some
    operation.
    By connecting multiple nodes' inputs and outputs together, computational graphs can
    be formed.
    These can be collected under a parent, such that new graphs can be composed of
    one or more sub-graphs.

    This is an abstract class.
    Children *must* define how :attr:`inputs` and :attr:`outputs` are constructed,
    what will happen :meth:`_on_run`, the :attr:`run_args` that will get passed to
    :meth:`_on_run`, and how to :meth:`process_run_result` once :meth:`_on_run` finishes.
    They may optionally add additional signal channels to the signals IO.

    Attributes:
        failed (bool): Whether the node raised an error calling :meth:`run`. (Default
            is False.)
        future (concurrent.futures.Future | None): A futures object, if the node is
            currently running or has already run using an executor.
        label (str): A name for the node.
        parent (pyiron_workflow.composite.Composite | None): The parent object
            owning this, if any.
        recovery: (BackendIdentifier | StorageInterface | None): The storage
            backend to use for saving a "recovery" file if the node execution crashes
            and this is the parent-most node. Default is `"pickle"`, setting `None`
            will prevent any file from being saved.
        running (bool): Whether the node has called :meth:`run` and has not yet
            received output from this call. (Default is False.)
        checkpoint (BackendIdentifier | StorageInterface | None): Whether to trigger a
            save of the entire graph after each run of the node, and if so what storage
            back end to use. (Default is None, don't do any checkpoint saving.)
        use_cache (bool): Whether or not to cache the inputs and, when the current
            inputs match the cached input (by `==` comparison), to bypass running the
            node and simply continue using the existing outputs. Note that you may be
            able to trigger a false cache hit in some special case of non-idempotent
            nodes working on mutable data.
    """

    use_cache = True

    def __init__(
        self,
        *args,
        label: str | None = None,
        parent: Composite | None = None,
        delete_existing_savefiles: bool = False,
        autoload: BackendIdentifier | StorageInterface | None = None,
        autorun: bool = False,
        checkpoint: BackendIdentifier | StorageInterface | None = None,
        **kwargs,
    ):
        """
        A parent class for objects that can form nodes in the graph representation of a
        computational workflow.

        Initialization ends with a routine :meth:`_after_node_setup` that may,
        depending on instantiation arguments, try to actually execute the node. Since
        child classes may need to get things done before this point, we want to make
        sure that this happens _after_ all the other setup. This can be accomplished
        by children (a) sticking stuff that is independent of `super().__init__` calls
        before the super call, and (b) overriding :meth:`_setup_node(self)` to do any
        remaining, parameter-free setup. This latter function gets called prior to any
        execution.

        Initialization will also try to parse any outstanding `args` and `kwargs` as
        input to the node's input channels. For node class developers, that means it's
        also important that `Node` parentage appear to the right-most of the
        inheritance set in the class definition, so that it's invokation of `__init__`
        appears as late as possible with the minimal set of args and kwargs.

        Args:
            label (str): A name for this node.
            *args: Interpreted as node input data, in order of input channels.
            parent: (Composite|None): The composite node that owns this as a child.
            delete_existing_savefiles (bool): Whether to look for and delete any
                matching save files at instantiation. Uses all default storage
                back ends and anything passed to :param:`autoload`. (Default is False,
                leave those files alone!)
            autoload (BackendIdentifier | StorageInterface | None): The back end
                to use for checking whether node data can be loaded from file. A None
                value indicates no auto-loading. (Default is "pickle".)
            autorun (bool): Whether to run at the end of initialization.
            checkpoint (BackendIdentifier | StorageInterface | None): The storage
                back end to use for saving the overall graph at the end of this node's
                run. (Default is None, don't do checkpoint saves.)
            **kwargs: Interpreted as node input data, with keys corresponding to
                channel labels.
        """
        super().__init__(label=label, parent=parent)

        self._signals = Signals()
        self._signals.input.run = InputSignal("run", self, self.run)
        self._signals.input.accumulate_and_run = AccumulatingInputSignal(
            "accumulate_and_run", self, self.run
        )
        self._signals.output.ran = OutputSignal("ran", self)
        self._signals.output.failed = OutputSignal("failed", self)

        self.checkpoint = checkpoint
        self.recovery: BackendIdentifier | StorageInterface | None = "pickle"
        self._remove_executorlib_cache: bool = True  # Power-user override for cleaning
        # up temporary serialized results from runs with executorlib; intended to be
        # used for testing
        self._cached_inputs: dict[str, Any] | None = None

        self._user_data: dict[str, Any] = {}
        # A place for power-users to bypass node-injection

        self._setup_node()
        self._after_node_setup(
            *args,
            delete_existing_savefiles=delete_existing_savefiles,
            autoload=autoload,
            autorun=autorun,
            **kwargs,
        )

    def _setup_node(self) -> None:
        """
        Called _before_ :meth:`Node.__init__` finishes.

        Child node classes can use this for any parameter-free node setup that should
        happen _before_ :meth:`Node._after_node_setup` gets called.
        """

    def _after_node_setup(
        self,
        *args,
        delete_existing_savefiles: bool = False,
        autoload: BackendIdentifier | StorageInterface | None = None,
        autorun: bool = False,
        **kwargs,
    ):
        if delete_existing_savefiles:
            self.delete_storage(backend=autoload)

        if autoload is not None:
            for backend in available_backends(backend=autoload):
                if backend.has_saved_content(self):
                    logger.info(
                        f"A saved file was found for the node {self.full_label} -- "
                        f"attempting to load it...(To delete the saved file instead, "
                        f"use `delete_existing_savefiles=True`) "
                    )
                    self.load(backend=autoload)
                    break

        self.set_input_values(*args, **kwargs)

        if autorun:
            with contextlib.suppress(ReadinessError):
                self.run()

    @property
    def channel(self) -> OutputDataWithInjection:
        """
        The single output channel. Fulfills the interface expectations for the
        :class:`HasChannel` mixin and allows this object to be used directly for
        forming connections, etc.

        Returns:
            (OutputDataWithInjection): The single output channel.

        Raises:
            AmbiguousOutputError: If there is not exactly one output channel.
        """
        if len(self.outputs) != 1:
            raise AmbiguousOutputError(
                f"Tried to access the channel value of {self.label}, but this is only "
                f"possible when there is a single output channel -- {self.label} has: "
                f"{self.outputs.labels}. Access probably occurred attempting to use "
                f"this object like an output channel, e.g. with injection or to form a "
                f"connection. Either make sure it has exactly one output channel, or "
                f"use the particular channel you want directly."
            )
        else:
            return self.outputs[self.outputs.labels[0]]

    @property
    def graph_path(self) -> str:
        """
        The path of node labels from the graph root (parent-most node in this lexical
        path) down to this node.
        """
        prefix = self.parent.lexical_path if isinstance(self.parent, Node) else ""
        return prefix + self.lexical_delimiter + self.label

    @property
    def graph_root(self) -> Node:
        """The parent-most node in this lexical path."""
        return self.parent.graph_root if isinstance(self.parent, Node) else self

    @property
    @abstractmethod
    def inputs(self) -> Inputs: ...

    @property
    @abstractmethod
    def outputs(self) -> OutputsWithInjection: ...

    @property
    def signals(self) -> Signals:
        """
        A container for input and output signals, which are channels for controlling
        execution flow. By default, has a :attr:`signals.inputs.run` channel which has
        a callback to the :meth:`run` method that fires whenever _any_ of its
        connections sends a signal to it, a :attr:`signals.inputs.accumulate_and_run`
        channel which has a callback to the :meth:`run` method but only fires after
        _all_ its connections send at least one signal to it, and `signals.outputs.ran`
        which gets called when the `run` method is finished.

        Additional signal channels in derived classes can be added to
        :attr:`signals.inputs` and  :attr:`signals.outputs` after this mixin class is
        initialized.
        """
        return self._signals

    @property
    def connected(self) -> bool:
        """Whether _any_ of the IO (including signals) are connected."""
        return self.inputs.connected or self.outputs.connected or self.signals.connected

    @property
    def fully_connected(self) -> bool:
        """Whether _all_ of the IO (including signals) are connected."""
        return (
            self.inputs.fully_connected
            and self.outputs.fully_connected
            and self.signals.fully_connected
        )

    def disconnect(self) -> list[tuple[Channel, Channel]]:
        """
        Disconnect all connections belonging to inputs, outputs, and signals channels.

        Returns:
            [list[tuple[Channel, Channel]]]: A list of the pairs of channels that no
                longer participate in a connection.
        """
        destroyed_connections = []
        destroyed_connections.extend(self.inputs.disconnect())
        destroyed_connections.extend(self.outputs.disconnect())
        destroyed_connections.extend(self.signals.disconnect())
        return destroyed_connections

    def activate_strict_hints(self) -> None:
        """Enable type hint checks for all data IO"""
        self.inputs.activate_strict_hints()
        self.outputs.activate_strict_hints()

    def deactivate_strict_hints(self) -> None:
        """Disable type hint checks for all data IO"""
        self.inputs.deactivate_strict_hints()
        self.outputs.deactivate_strict_hints()

    def _connect_output_signal(self, signal: OutputSignal) -> None:
        self.signals.input.run.connect(signal)

    def __rshift__(self, other: InputSignal | Node) -> InputSignal | Node:
        """
        Allows users to connect run and ran signals like: `first >> second`.
        """
        other._connect_output_signal(self.signals.output.ran)
        return other

    def _connect_accumulating_input_signal(
        self, signal: AccumulatingInputSignal
    ) -> None:
        self.signals.output.ran.connect(signal)

    def __lshift__(self, others: tuple[OutputSignal | Node, ...]):
        """
        Connect one or more `ran` signals to `accumulate_and_run` signals like:
        `this << some_object, another_object, or_by_channel.signals.output.ran`
        """
        self.signals.input.accumulate_and_run << others

    def _get_complete_input(self, *args, **kwargs) -> dict[str, Any]:
        if len(args) > len(self.inputs.labels):
            raise ValueError(
                f"Received {len(args)} args, but only have {len(self.inputs.labels)} "
                f"input channels available"
            )
        keyed_args = dict(zip(self.inputs.labels, args, strict=False))

        if len(set(keyed_args.keys()).intersection(kwargs.keys())) > 0:
            raise ValueError(
                f"n args are interpreted using the first n input channels "
                f"({self.inputs.labels}), but this conflicted with received kwargs "
                f"({list(kwargs.keys())}) -- perhaps the input was ordered differently "
                f"than expected? Got args {args} and kwargs {kwargs}."
            )

        kwargs.update(keyed_args)

        self._ensure_all_input_keys_present(kwargs.keys(), self.inputs.labels)

        return kwargs

    def set_input_values(self, *args, **kwargs) -> None:
        """
        Match keywords to input channels and update their values.

        Throws a warning if a keyword is provided that cannot be found among the input
        keys.

        Args:
            *args: values assigned to inputs in order of appearance.
            **kwargs: input key - input value (including channels for connection) pairs.

        Raises:
            (ValueError): If more args are received than there are inputs available.
            (ValueError): If there is any overlap between channels receiving values
                from `args` and those from `kwargs`.
            (ValueError): If any of the `kwargs` keys do not match available input
                labels.
        """
        for k, v in self._get_complete_input(*args, **kwargs).items():
            self.inputs[k] = v

    @staticmethod
    def _ensure_all_input_keys_present(used_keys, available_keys):
        diff = set(used_keys).difference(available_keys)
        if len(diff) > 0:
            raise ValueError(
                f"{diff} not found among available inputs: {available_keys}"
            )

    @property
    def _owned_io_panels(self) -> list[IO]:
        return [
            self.inputs,
            self.outputs,
            self.signals.input,
            self.signals.output,
        ]

    def data_input_locked(self):
        return self.running

    @property
    def _readiness_dict(self) -> dict[str, bool]:
        dict = super()._readiness_dict
        for k, v in self.inputs.items():
            dict[f"inputs.{k}"] = v.ready
        return dict

    @property
    def _readiness_error_message(self) -> str:
        return (
            f"{self.label} received a run command but is not ready. The node "
            f"should be neither running nor failed, and all input values should"
            f" conform to type hints.\n" + self.readiness_report
        )

    def _is_using_wrapped_excutorlib_executor(self) -> bool:
        return self.executor is not None and (
            isinstance(self.executor, CacheOverride)
            or (
                isinstance(self.executor, tuple)
                and isinstance(self.executor[0], type)
                and issubclass(self.executor[0], CacheOverride)
            )
        )

    def _clean_wrapped_executorlib_executor_cache(self) -> None:
        self._wrapped_executorlib_cache_file.unlink()
        cache_subdir = self.as_path() / CacheOverride.override_cache_file_name
        if pathlib.Path(cache_subdir).is_dir():
            shutil.rmtree(cache_subdir)
        self.clean_path()

    @property
    def _wrapped_executorlib_cache_file(self) -> Path:
        """For internal use to clean up cached executorlib files"""
        # Depends on executorlib implementation details not protected by semver
        file_name = CacheOverride.override_cache_file_name + "_o.h5"
        return self.as_path() / file_name

    def on_run(self, *args, **kwargs) -> Any:
        return self._on_run(*args, **kwargs)

    @abstractmethod
    def _on_run(self, *args, **kwargs) -> Any: ...

    def run(
        self,
        *args,
        run_data_tree: bool = False,
        run_parent_trees_too: bool = False,
        fetch_input: bool = True,
        check_readiness: bool = True,
        raise_run_exceptions: bool = True,
        rerun: bool = False,
        emit_ran_signal: bool = True,
        **kwargs,
    ):
        """
        The master method for running in a variety of ways.
        By default, whatever data is currently available in upstream nodes will be
        fetched, if the input all conforms to type hints then this node will be run
        (perhaps using an executor), and  finally the `ran` signal will be emitted to
        trigger downstream runs.

        If executor information is specified, execution happens on that process, a
        callback is registered, and futures object is returned.

        Input values can be updated at call time with kwargs, but this happens _first_
        so any input updates that happen as a result of the computation graph will
        override these by default. If you really want to execute the node with a
        particular set of input, set it all manually and use `execute` (or `run` with
        carefully chosen flags).

        Args:
            run_data_tree (bool): Whether to first run all upstream nodes in the data
                graph. (Default is False.)
            run_parent_trees_too (bool): Whether to recursively run the data tree in
                parent nodes (if any). (Default is False.)
            fetch_input (bool): Whether to first update inputs with the
                highest-priority connections holding data (i.e. the first valid
                connection; and the most recently formed connections appear first
                unless the connections list has been manually tampered with). (Default
                is True.)
            check_readiness (bool): Whether to raise an exception if the node is not
                :attr:`ready` to run after fetching new input. (Default is True.)
            raise_run_exceptions (bool): Whether to raise exceptions encountered during
                the run, or just ignore them. (Default is True, raise them!)
            rerun (bool): Whether to force-set :attr:`running` and :attr:`failed` to
                `False` before running. (Default is False.)
            emit_ran_signal (bool): Whether to fire off all the output `ran` signal
                afterwards. (Default is True.)
            **kwargs: Keyword arguments matching input channel labels; used to update
                the input channel values before running anything.

        Returns:
            (Any | Future): The result of running the node, or a futures object (if
                running on an executor).

        Note:
            Running data trees is a pull-based paradigm and only compatible with graphs
            whose data forms a directed acyclic graph (DAG).

        Note:
            Kwargs updating input channel values happens _first_ and will get
            overwritten by any subsequent graph-based data manipulation.
        """

        return super().run(
            check_readiness=check_readiness,
            raise_run_exceptions=raise_run_exceptions,
            rerun=rerun,
            before_run_kwargs={
                "run_data_tree": run_data_tree,
                "run_parent_trees_too": run_parent_trees_too,
                "fetch_input": fetch_input,
                "emit_ran_signal": emit_ran_signal,
                "input_args": args,
                "input_kwargs": kwargs,
            },
            run_finally_kwargs={
                "emit_ran_signal": emit_ran_signal,
                "raise_run_exceptions": raise_run_exceptions,
            },
        )

    def _before_run(
        self,
        /,
        check_readiness: bool,
        rerun: bool,
        run_data_tree: bool,
        run_parent_trees_too: bool,
        fetch_input: bool,
        emit_ran_signal: bool,
        input_args: tuple[Any, ...],
        input_kwargs: dict[str, Any],
    ) -> tuple[bool, Any]:
        if self.running:
            if len(input_args) > 0 or len(input_kwargs) > 0:
                raise InputLockedError(
                    f"Node {self.label} is running. Input values are not allowed to be "
                    f"updated in this state, but got args {input_args} and kwargs "
                    f"{input_kwargs}."
                )
            if self.future is not None:
                if rerun:
                    raise WaitingForFutureError(
                        f"Node {self.label} is running and has a future attached to "
                        f"it. It cannot be rerun in this state."
                    )
                else:
                    return True, self.future
            if self._is_using_wrapped_excutorlib_executor():
                return False, None  # Let it cook
            elif not rerun:
                raise ReadinessError(self._readiness_error_message)

        if self.failed and check_readiness and not rerun:
            raise ReadinessError(self._readiness_error_message)

        if run_data_tree:
            self.run_data_tree(run_parent_trees_too=run_parent_trees_too)

        self.set_input_values(*input_args, **input_kwargs)
        if fetch_input:
            self.inputs.fetch()

        if self.use_cache and self.cache_hit:  # Read and use cache
            self._on_cache_hit()
            if (self.parent is None or not self.parent.running) and emit_ran_signal:
                self.emit()
            elif self.parent is not None and self.parent.running:
                self.parent.register_child_starting(self)
                self.parent.register_child_finished(self)
                if emit_ran_signal:
                    self.parent.register_child_emitting(self)
            return True, self._outputs_to_run_return()
        else:
            self._on_cache_miss()
            if self.use_cache:  # Write cache and continue
                self._cache_inputs()

        return super()._before_run(check_readiness=check_readiness, rerun=rerun)

    def _on_cache_hit(self) -> None:
        """A hook for subclasses to act on cache hits"""
        return

    def _on_cache_miss(self) -> None:
        """A hook for subclasses to act on cache misses"""
        return

    def _cache_inputs(self):
        self._cached_inputs = self.inputs.to_value_dict()

    def clear_cache(self):
        self._cached_inputs = None

    def _run(
        self,
        raise_run_exceptions: bool,
        run_exception_kwargs: dict,
        run_finally_kwargs: dict,
        finish_run_kwargs: dict,
    ) -> Any | tuple | Future:
        if self.parent is not None and self.parent.running:
            self.parent.register_child_starting(self)
        return super()._run(
            raise_run_exceptions=raise_run_exceptions,
            run_exception_kwargs=run_exception_kwargs,
            run_finally_kwargs=run_finally_kwargs,
            finish_run_kwargs=finish_run_kwargs,
        )

    def _run_finally(self, /, emit_ran_signal: bool, raise_run_exceptions: bool):
        super()._run_finally()
        if self.parent is not None and self.parent.running:
            self.parent.register_child_finished(self)
        if self.checkpoint is not None:
            self.save_checkpoint(self.checkpoint)

        if emit_ran_signal:
            if self.parent is None or not self.parent.running:
                self.emit()
            else:
                self.parent.register_child_emitting(self)

        if (
            self.failed
            and raise_run_exceptions
            and self.recovery is not None
            and self.graph_root is self
        ):
            self.save(
                backend=self.recovery, filename=self.as_path().joinpath("recovery")
            )

        if (
            self._remove_executorlib_cache
            and self._is_using_wrapped_excutorlib_executor()
        ):
            self._clean_wrapped_executorlib_executor_cache()

    def run_data_tree(self, run_parent_trees_too=False) -> None:
        """
        Use topological analysis to build a tree of all upstream dependencies and run
        them.

        Args:
            run_parent_trees_too (bool): First, call the same method on this node's
                parent (if one exists), and recursively up the parentage tree. (Default
                is False, only run nodes in this scope, i.e. sharing the same parent.)
        """
        if run_parent_trees_too and self.parent is not None:
            self.parent.run_data_tree(run_parent_trees_too=True)
            self.parent.inputs.fetch()

        data_tree_nodes = get_nodes_in_data_tree(self)
        # If we have a parent, delegate to it
        if self.parent is not None:
            self.parent.run_data_tree_for_child(self)
            return

        # The rest of this method handles the case when self.parent is None
        label_map = {}
        nodes = {}

        for node in data_tree_nodes:
            if node.executor is not None:
                raise ValueError(
                    f"Running the data tree is pull-paradigm action, and is "
                    f"incompatible with using executors. While running "
                    f"{self.full_label}, an executor request was found on "
                    f"{node.full_label}"
                )

        for node in data_tree_nodes:
            modified_label = node.label + str(id(node))
            label_map[modified_label] = node.label
            node.label = modified_label  # Ensure each node has a unique label
            # This is necessary when the nodes do not have a workflow and may thus have
            # arbitrary labels.
            # This is pretty ugly; it would be nice to not depend so heavily on labels.
            # Maybe we could switch a bunch of stuff to rely on the unique ID?
            nodes[modified_label] = node

        try:
            disconnected_pairs, starters = set_run_connections_according_to_linear_dag(
                nodes
            )
            data_tree_starters = list(set(starters).intersection(data_tree_nodes))
        except Exception as e:
            # If the dag setup fails it will repair any connections it breaks before
            # raising the error, but we still need to repair our label changes
            for modified_label, node in nodes.items():
                node.label = label_map[modified_label]
            raise e

        try:
            if len(data_tree_starters) > 1 or data_tree_starters[0] is not self:
                self.signals.disconnect_run()
                # Don't let anything upstream trigger _this_ node

                for starter in data_tree_starters:
                    starter.run()  # Now push from the top
            # Otherwise the requested node is the only one in the data tree, so there's
            # nothing upstream to run.
        finally:
            # No matter what, restore the original connections and labels afterwards
            for modified_label, node in nodes.items():
                node.label = label_map[modified_label]
                node.signals.disconnect_run()
            for c1, c2 in disconnected_pairs:
                c1.connect(c2)

    @property
    def cache_hit(self) -> bool:
        try:
            return self.inputs.to_value_dict() == self._cached_inputs
        except Exception:
            return False

    def _outputs_to_run_return(self):
        return DotDict(self.outputs.to_value_dict())

    @property
    def emitting_channels(self) -> tuple[OutputSignal, ...]:
        if self.failed:
            return (self.signals.output.failed,)
        else:
            return (self.signals.output.ran,)

    def emit(self):
        for channel in self.emitting_channels:
            channel()

    def execute(self, *args, **kwargs):
        """
        A shortcut for :meth:`run` with particular flags.

        Run the node with whatever input it currently has (or is given as kwargs here),
        run it on this python process, and don't emit the `ran` signal afterwards.

        Intended to be useful for debugging by just forcing the node to do its thing
        right here, right now, and as-is.
        """
        return self.run(
            *args,
            run_data_tree=False,
            run_parent_trees_too=False,
            fetch_input=False,
            check_readiness=False,
            rerun=False,
            emit_ran_signal=False,
            **kwargs,
        )

    def pull(self, *args, run_parent_trees_too=False, **kwargs):
        """
        A shortcut for :meth:`run` with particular flags.

        Runs nodes upstream in the data graph, then runs this node without triggering
        any downstream runs. By default only runs sibling nodes, but can optionally
        require the parent node to pull in its own upstream runs (this is recursive
        up to the parent-most object).

        Args:
            run_parent_trees_too (bool): Whether to (recursively) require the parent to
                first pull.
        """
        return self.run(
            *args,
            run_data_tree=True,
            run_parent_trees_too=run_parent_trees_too,
            fetch_input=True,
            check_readiness=True,
            rerun=False,
            emit_ran_signal=False,
            **kwargs,
        )

    def push(self, *args, **kwargs):
        """
        Exactly like :meth:`run` with all the same flags, _except_ it handles an edge
        case where you are trying to directly run the child node of a
        :class:`pyiron_workflow.workflow.Workflow` before it has had any chance to
        configure its execution signals.
        _If_ the parent is a workflow set up to automate execution flow, does that
        _first_ then runs as usual.
        """
        # Alright, time for more egregious hacking
        # Normally, running will work in a push-like mode _BUT_, because Workflow's are
        # a flexible dynamic thing, they normally construct their execution signals on
        # the fly at each run invocation. This is not the case for Macros, where the
        # run configuration, if automated at all, happens right at macro instantiation.
        # So there's this horrible edge case where you build a workflow, then
        # immediately try to run one of its children directly, naively expecting that
        # the run will push downstream executions like it does in a macro -- except it
        # _doesn't_ because there are _no signal connections at all yet!_
        # Building these on _every_ run would be needlessly expensive, so this method
        # exists as a hacky guaranteed way to secure push-like run behaviour regardless
        # of the context you're calling from.
        if self.parent is not None:
            return self.parent.push_child(self, *args, **kwargs)
        else:
            return self.run(*args, **kwargs)

    def __call__(self, *args, **kwargs) -> None:
        """
        A shortcut for :meth:`pull` that automatically runs the entire set of upstream data
        dependencies all the way to the parent-most graph object.
        """
        return self.pull(*args, run_parent_trees_too=True, **kwargs)

    @property
    def ready(self) -> bool:
        """
        Whether the inputs are all ready and the node is neither already running nor
        already failed.
        """
        return super().ready and self.inputs.ready

    @property
    def color(self) -> str:
        """A hex code color for use in drawing."""
        return SeabornColors.white

    def draw(
        self,
        depth: int = 1,
        rankdir: Literal["LR", "TB"] = "LR",
        size: tuple | None = None,
        save: bool = False,
        view: bool = False,
        directory: Path | str | None = None,
        filename: Path | str | None = None,
        format: str | None = None,
        cleanup: bool = True,
    ) -> graphviz.graphs.Digraph:
        """
        Draw the node structure and return it as a graphviz object.

        A selection of the :func:`graphviz.Graph.render` method options are exposed,
        and if :param:`view` or :param:`filename` is provided, this will be called
        before returning the graph.
        The graph file and rendered image will be stored in a directory based of the
        node's lexical path, unless a :param:`directory` is explicitly set.
        This is purely for convenience -- since we directly return a graphviz object
        you can instead use this to leverage the full power of graphviz.

        Args:
            depth (int): How deeply to decompose the representation of composite nodes
                to reveal their inner structure. (Default is 1, which will show owned
                nodes if _this_ is a composite node, but all children will be drawn
                at the level of showing their IO only.) A depth value greater than the
                max depth of the node will have no adverse side effects.
            rankdir ("LR" | "TB"): Use left-right or top-bottom graphviz `rankdir` to
                orient the flow of the graph.
            size (tuple[int | float, int | float] | None): The size of the diagram, in
                inches(?); respects ratio by scaling until at least one dimension
                matches the requested size. (Default is None, automatically size.)
            save (bool): Render the graph image. (Default is False. When True, all
                other defaults will yield a PDF in the node's working directory.)
            view (bool): `graphviz.Graph.render` argument, open the rendered result
                with the default application. (Default is False. When True, default
                values for the directory and filename are supplied by the node working
                directory and label.)
            directory (Path|str|None): `graphviz.Graph.render` argument, (sub)directory
                for source saving and rendering. (Default is None, which uses the
                node's working directory.)
            filename (Path|str): `graphviz.Graph.render` argument, filename for saving
                the source. (Default is None, which uses the node label + `"_graph"`.
            format (str|None): `graphviz.Graph.render` argument, the output format used
                for rendering ('pdf', 'png', etc.).
            cleanup (bool): `graphviz.Graph.render` argument, delete the source file
                after successful rendering. (Default is True -- unlike graphviz.)

        Returns:
            (graphviz.graphs.Digraph): The resulting graph object.

        """
        size_str = f"{size[0]},{size[1]}" if size is not None else None
        graph = GraphvizNode(self, depth=depth, rankdir=rankdir, size=size_str).graph
        if save or view or filename is not None:
            directory = self.as_path() if directory is None else Path(directory)
            filename = self.label + "_graph" if filename is None else filename
            graph.render(
                view=view,
                directory=directory,
                filename=filename,
                format=format,
                cleanup=cleanup,
            )
        return graph

    def __str__(self):
        return (
            f"{self.label} ({self.__class__.__name__}):\n"
            f"{str(self.inputs)}\n"
            f"{str(self.outputs)}\n"
            f"{str(self.signals)}"
        )

    _save_load_warnings = """
        HERE BE DRAGONS!!!

        Warning:
            This almost certainly only fails for subclasses of :class:`Node` that don't
            override `node_function` or `macro_creator` directly, as these are expected 
            to be part of the class itself (and thus already present on our instantiated 
            object) and are never stored. Nodes created using the provided decorators 
            should all work.

        Warning:
            If you modify a `Macro` class in any way (changing its IO maps, rewiring 
            internal connections, or replacing internal nodes), don't expect 
            saving/loading to work.

        Warning:
            If the underlying source code has changed since saving (i.e. the node doing 
            the loading does not use the same code as the node doing the saving, or the 
            nodes in some node package have been modified), then all bets are off.

    """

    def save(
        self,
        backend: BackendIdentifier | StorageInterface = "pickle",
        filename: str | Path | None = None,
        **kwargs,
    ):
        """
        Writes the node to file using the requested interface as a back end.

        Args:
            backend (str | StorageInterface): The interface to use for serializing the
                node. (Default is "pickle", which loads the standard pickling back end.)
            filename (str | Path | None): The name of the file (without extensions) at
                which to save the node. (Default is None, which uses the node's
                lexical path.)
            **kwargs: Back end-specific keyword arguments.
        """
        for selected_backend in available_backends(
            backend=backend, only_requested=True
        ):
            selected_backend.save(node=self, filename=filename, **kwargs)

    save.__doc__ = cast(str, save.__doc__) + _save_load_warnings

    def save_checkpoint(self, backend: BackendIdentifier | StorageInterface = "pickle"):
        """
        Triggers a save on the parent-most node.

        Args:
            backend (str | StorageInterface): The interface to use for serializing the
                node. (Default is "pickle", which loads the standard pickling back end.)
        """
        self.graph_root.save(backend=backend)

    def load(
        self,
        backend: BackendIdentifier | StorageInterface = "pickle",
        only_requested=False,
        filename: str | Path | None = None,
        **kwargs,
    ):
        """

        Loads the node file and set the loaded state as the node's own.

        Args:
            backend (str | StorageInterface): The interface to use for serializing the
                node. (Default is "pickle", which loads the standard pickling back end.)
            only_requested (bool): Whether to _only_ try loading from the specified
                backend, or to loop through all available backends. (Default is False,
                try to load whatever you can find.)
            filename (str | Path | None): The name of the file (without extensions) at
                which to save the node. (Default is None, which uses the node's
                lexical path.)
            **kwargs: back end-specific arguments (only likely to work in combination
                with :param:`only_requested`, otherwise there's nothing to be specific
                _to_.)

        Raises:
            FileNotFoundError: when nothing got loaded.
            TypeError: when the saved node has a different class name.
        """
        if self.running:
            raise ValueError(
                "Cannot load a node while it is running. If you are sure loading now "
                "is the correct thing to do, you can set `self.running=True` where "
                "`self` is this node object."
            )
        for selected_backend in available_backends(
            backend=backend, only_requested=only_requested
        ):
            inst = selected_backend.load(
                node=self if filename is None else None, filename=filename, **kwargs
            )
            if inst is not None:
                break
        if inst is None:
            raise FileNotFoundError(f"{self.label} could not find saved content.")

        if inst.__class__ != self.__class__:
            raise TypeError(
                f"{self.label} cannot load, as it has type "
                f"{self.__class__.__name__},  but the saved node has type "
                f"{inst.__class__.__name__}"
            )
        self.__setstate__(inst.__getstate__())

    load.__doc__ = cast(str, load.__doc__) + _save_load_warnings

    def delete_storage(
        self,
        backend: BackendIdentifier | StorageInterface | None = None,
        only_requested: bool = False,
        filename: str | Path | None = None,
        *,
        delete_even_if_not_empty: bool = False,
        **kwargs,
    ):
        """
        Remove save file(s).

        Args:
            backend (str | StorageInterface): The interface to use for serializing the
                node. (Default is "pickle", which loads the standard pickling back end.)
            only_requested (bool): Whether to _only_ try loading from the specified
                backend, or to loop through all available backends. (Default is False,
                try to load whatever you can find.)
            filename (str | Path | None): The name of the file (without extensions) at
                which to save the node. (Default is None, which uses the node's
                lexical path.)
            delete_even_if_not_empty (bool): Whether to delete the file even if it is
                not empty. (Default is False, which will only delete the file if it is
                empty, i.e. has no content in it.)
            **kwargs: back end-specific arguments (only likely to work in combination
                with :param:`only_requested`, otherwise there's nothing to be specific
                _to_.)
        """
        for selected_backend in available_backends(
            backend=backend, only_requested=only_requested
        ):
            selected_backend.delete(
                node=self if filename is None else None,
                filename=filename,
                delete_even_if_not_empty=delete_even_if_not_empty,
                **kwargs,
            )

    def has_saved_content(
        self,
        backend: BackendIdentifier | StorageInterface | None = None,
        only_requested: bool = False,
        filename: str | Path | None = None,
        **kwargs,
    ):
        """
        Whether any save files can be found at the canonical location for this node.

        Args:
            backend (str | StorageInterface): The interface to use for serializing the
                node. (Default is "pickle", which loads the standard pickling back end.)
            only_requested (bool): Whether to _only_ try loading from the specified
                backend, or to loop through all available backends. (Default is False,
                try to load whatever you can find.)
            filename (str | Path | None): The name of the file (without extensions) at
                which to save the node. (Default is None, which uses the node's
                lexical path.)
            **kwargs: back end-specific arguments (only likely to work in combination
                with :param:`only_requested`, otherwise there's nothing to be specific
                _to_.)

        Returns:
            bool: Whether any save files were found
        """
        return any(
            be.has_saved_content(
                node=self if filename is None else None, filename=filename, **kwargs
            )
            for be in available_backends(backend=backend, only_requested=only_requested)
        )

    @property
    def import_ready(self) -> bool:
        """
        Checks whether `importlib` can find this node's class, and if so whether the
        imported object matches the node's type.

        Returns:
            (bool): Whether the imported module and name of this node's class match
                its type.
        """
        try:
            module = self.__class__.__module__
            class_ = getattr(import_module(module), self.__class__.__name__)
            if module == "__main__":
                logger.warning(f"{self.label} is only defined in __main__")
            return type(self) is class_
        except (ModuleNotFoundError, AttributeError):
            return False

    @property
    def import_readiness_report(self):
        print(self.report_import_readiness())

    def report_import_readiness(self, tabs=0, report_so_far=""):
        newline = "\n" if len(report_so_far) > 0 else ""
        tabspace = tabs * "\t"
        return (
            report_so_far + f"{newline}{tabspace}{self.label}: "
            f"{'ok' if self.import_ready else 'NOT IMPORTABLE'}"
        )

    def display_state(self, state=None, ignore_private=True):
        state = dict(self.__getstate__()) if state is None else state
        if self.parent is not None:
            state["parent"] = self.parent.full_label
        if len(state["_user_data"]) > 0:
            self._make_entry_public(state, "_user_data", "user_data")
        return super().display_state(state=state, ignore_private=ignore_private)

    @classmethod
    def _extra_info(cls) -> str:
        """
        Any additional info that may be particularly useful for users of the node class.
        """
        return ""



"""
For interface specification.

There are cases where we want to be able to depend on the presence of a particular
attribute or method, but care very little about its details -- nothing beyond, perhaps,
type hints. These mixins allow us to guarantee the presence of such interfaces while
leaving their actual implementation up to other classes in order to have the weakest
possible coupling between different components of a composed class.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from pyiron_workflow.channels import Channel


class UsesState:
    """
    A mixin for any class using :meth:`__getstate__` or :meth:`__setstate__`.

    Guarantees that `super()` can always be called in these methods to return a copy
    of the state dict or to update it, respectively.
    """

    def __getstate__(self):
        # Make a shallow(! careful!) copy of the state so any modifications don't
        # immediately impact the object we're getting the state from)
        return dict(self.__dict__)

    def __setstate__(self, state):
        self.__dict__.update(**state)


class HasLabel(ABC):
    """
    A mixin to guarantee the label interface exists.
    """

    _label: str

    @property
    def label(self) -> str:
        """A label for the object."""
        return self._label

    @label.setter
    def label(self, new_label: str):
        self._check_label(new_label)
        self._label = new_label

    def _check_label(self, new_label: str) -> None:
        """
        Extensible checking routine for label validity.
        """
        if not isinstance(new_label, str):
            raise TypeError(f"Expected a string label but got {new_label}")

    @property
    def full_label(self) -> str:
        """
        A more verbose label based off the underlying label attribute (and possibly
        other data) -- in the root class, it's just the same as the label.
        """
        return self.label


class HasChannel(ABC):
    """
    A mix-in class for use with the :class:`Channel` class and its children.

    This is useful for letting channels attempt to connect to non-channel objects
    directly by pointing them to some channel that object holds.
    """

    @property
    @abstractmethod
    def channel(self) -> Channel: ...


class HasRun(ABC):
    """
    A mixin to guarantee that the :meth:`run` method exists.
    """

    @abstractmethod
    def run(self, *args, **kwargs): ...



"""
A base class for macro nodes, which are composite like workflows but have a static
interface and are not intended to be internally modified after instantiation.
"""

from __future__ import annotations

import re
from abc import ABC, abstractmethod
from collections.abc import Callable
from inspect import getsource
from typing import TYPE_CHECKING, Self, TypeAlias

from pyiron_snippets.factory import classfactory

from pyiron_workflow.channels import OutputData
from pyiron_workflow.io import Inputs
from pyiron_workflow.mixin.injection import OutputsWithInjection
from pyiron_workflow.mixin.preview import ScrapesIO
from pyiron_workflow.nodes.composite import Composite
from pyiron_workflow.nodes.multiple_distpatch import dispatch_output_labels
from pyiron_workflow.nodes.static_io import StaticNode

if TYPE_CHECKING:
    from pyiron_workflow.channels import Channel


GraphOutput: TypeAlias = StaticNode | OutputData


class Macro(Composite, StaticNode, ScrapesIO, ABC):
    """
    A macro is a composite node that holds a graph with a fixed interface, like a
    pre-populated workflow that is the same every time you instantiate it.

    At instantiation, the macro uses a provided callable to build and wire the graph,
    then builds a static IO interface for this graph.
    This callable must use the macro object itself as the first argument (e.g. adding
    nodes to it).
    The provided callable may optionally specify further args and kwargs; these are
    used to pre-populate the macro with :class:`UserInput` nodes, although they may
    later be trimmed if the IO can be connected directly to child node IO without any
    loss of functionality.
    This can be especially helpful when more than one child node needs access to the
    same input value.
    Similarly, the callable may return any number of child nodes' output channels (or
    the node itself in the case of single-output nodes) as long as a commensurate
    number of labels for these outputs were provided to the class constructor.
    These function-like definitions of the graph creator callable can be used
    to build only input xor output, or both together.
    Macro input channel labels are scraped from the signature of the graph creator;
    for output, output labels can be provided explicitly as a class attribute or, as a
    fallback, they are scraped from the graph creator code return statement (stripping
    off the "{first argument}.", where {first argument} is whatever the name of the
    first argument is.

    Macro IO is _value linked_ to the child IO, so that their values stay synchronized,
    but the child nodes of a macro form an isolated sub-graph.

    As with function nodes, subclasses of :class:`Macro` may define a method for
    creating the graph.

    As with :class:`Workflow``, all DAG macros can determine their execution flow
    automatically, if you have cycles in your data flow, or otherwise want more control
    over the execution, all you need to do is specify the `node.signals.input.run`
    connections and :attr:`starting_nodes` list yourself.
    If only _one_ of these is specified, you'll get an error, but if you've provided
    both then no further checks of their validity/reasonableness are performed, so be
    careful.
    Unlike :class:`Workflow`, this execution flow automation is set up once at
    instantiation;
    If the macro is modified post-facto, you may need to manually re-invoke
    :meth:`configure_graph_execution`.

    Examples:
        Let's consider the simplest case of macros that just consecutively add 1 to
        their input:

        >>> from pyiron_workflow import as_macro_node, macro_node
        >>> def add_one(x):
        ...     result = x + 1
        ...     return result
        >>>
        >>> def add_three_macro(self, one__x):
        ...     self.one = self.create.function_node(add_one, x=one__x)
        ...     self.two = self.create.function_node(add_one, self.one)
        ...     self.three = self.create.function_node(add_one, self.two)
        ...     self.one >> self.two >> self.three
        ...     self.starting_nodes = [self.one]
        ...     return self.three

        In this case we had _no need_ to specify the execution order and starting nodes
        --it's just an extremely simple DAG after all! -- but it's done here to
        demonstrate the syntax.

        We can make a macro by passing this graph-building function (that takes a macro
        as its first argument, i.e. `self` from the macro's perspective) to the :class:`Macro`
        class. Then, we can use it like a regular node! Just like a workflow, the
        io is constructed from unconnected owned-node IO by combining node and channel
        labels.

        >>> macro = macro_node(add_three_macro, output_labels="three__result")
        >>> out = macro(one__x=3)
        >>> out.three__result
        6

        We can also nest macros, rename their IO, and provide access to
        internally-connected IO by inputs and outputs maps:

        >>> def nested_macro(self, inp):
        ...     self.a = self.create.function_node(add_one, x=inp)
        ...     self.b = self.create.macro_node(
        ...         add_three_macro, one__x=self.a, output_labels="three__result"
        ...     )
        ...     self.c = self.create.function_node(add_one, x=self.b)
        ...     return self.c, self.b
        >>>
        >>> macro = macro_node(
        ...     nested_macro, output_labels=("out", "intermediate")
        ... )
        >>> macro(inp=1)
        {'out': 6, 'intermediate': 5}

        Macros and workflows automatically generate execution flows when their data
        is acyclic.
        Let's build a simple macro with two independent tracks:

        >>> def modified_flow_macro(self, a__x=0, b__x=0):
        ...     self.a = self.create.function_node(add_one, x=a__x)
        ...     self.b = self.create.function_node(add_one, x=b__x)
        ...     self.c = self.create.function_node(add_one, x=self.b)
        ...     return self.a, self.c
        >>>
        >>> m = macro_node(modified_flow_macro, output_labels=("a", "c"))
        >>> m(a__x=1, b__x=2)
        {'a': 2, 'c': 4}

        We can override which nodes get used to start by specifying the
        :attr:`starting_nodes` property and (if necessary) reconfiguring the execution
        signals.
        Care should be taken here, as macro nodes may be creating extra input
        nodes that need to be considered.
        It's advisable to use :meth:`draw()` or to otherwise inspect the macro's
        children and their connections before manually updating execution flows.

        Let's use this and then observe how the `a` sub-node no longer gets run:

        >>> _ = m.disconnect_run()
        >>> m.starting_nodes = [m.b]
        >>> _ = m.b >> m.c
        >>> m(a__x=1000, b__x=2000)
        {'a': 2, 'c': 2002}

        (The `_` is just to catch and ignore output for the doctest, you don't
        typically need this.)

        Note how the `a` node is no longer getting run, so the output is not updated!
        Manually controlling execution flow is necessary for cyclic graphs (cf. the
        while loop meta-node), but best to avoid when possible as it's easy to miss
        intended connections in complex graphs.

        If there's a particular macro we're going to use again and again, we might want
        to consider making a new class for it using the decorator, just like we do for
        function nodes. If no output labels are explicitly provided as arguments to the
        decorator itself, these are scraped from the function return value, just like
        for function nodes (except the initial `macro` (or `self` or whatever the first
        argument is named) on any return values is ignored):

        >>> from pyiron_workflow.api import Macro
        >>> @Macro.wrap.as_macro_node
        ... def AddThreeMacro(self, x):
        ...     add_three_macro(self, one__x=x)
        ...     # We could also simply have decorated that function to begin with
        ...     return self.three
        >>>
        >>> macro = AddThreeMacro()
        >>> macro(x=0).three
        3

        Alternatively (and not recommended) is to make a new child class of
        :class:`Macro` that overrides the :meth:`graph_creator` arg such that
        the same graph is always created.

        >>> class AddThreeMacro(Macro):
        ...     _output_labels = ["three"]
        ...
        ...     def graph_creator(self, x):
        ...         add_three_macro(self, one__x=x)
        ...         return self.three
        >>>
        >>> macro = AddThreeMacro()
        >>> macro(x=0).three
        3

        We can also modify an existing macro at runtime by replacing nodes within it, as
        long as the replacement has fully compatible IO. There are three syntacic ways
        to do this. Let's explore these by going back to our `add_three_macro` and
        replacing each of its children with a node that adds 2 instead of 1.

        It's possible for the macro to hold nodes which are not publicly exposed for
        data and signal connections, but which will still internally execute and store
        data, e.g.:

        >>> @Macro.wrap.as_macro_node("lout", "n_plus_2")
        ... def LikeAFunction(self, lin: list,  n: int = 1):
        ...     self.plus_two = n + 2
        ...     self.sliced_list = lin[n:self.plus_two]
        ...     self.double_fork = 2 * n
        ...     return self.sliced_list, self.plus_two.channel
        >>>
        >>> like_functions = LikeAFunction(lin=[1,2,3,4,5,6], n=3)
        >>> sorted(like_functions().items())
        [('lout', [4, 5]), ('n_plus_2', 5)]

        >>> like_functions.double_fork.value
        6


    """

    def _setup_node(self) -> None:
        super()._setup_node()

        ui_nodes = self._prepopulate_ui_nodes_from_graph_creator_signature()
        returned_has_channel_objects = self.graph_creator(*ui_nodes)
        if returned_has_channel_objects is None:
            returned_has_channel_objects = ()
        elif isinstance(returned_has_channel_objects, GraphOutput):
            returned_has_channel_objects = (returned_has_channel_objects,)

        for node in ui_nodes:
            self.inputs[node.label].value_receiver = node.inputs.user_input

        for graph_output, output_channel_label in zip(
            returned_has_channel_objects,
            () if self._output_labels is None else self._output_labels,
            strict=False,
        ):
            graph_output.channel.value_receiver = self.outputs[output_channel_label]

        remaining_ui_nodes = self._purge_single_use_ui_nodes(ui_nodes)
        self._configure_graph_execution(remaining_ui_nodes)

    @abstractmethod
    def graph_creator(
        self: Self, *args, **kwargs
    ) -> GraphOutput | tuple[GraphOutput, ...] | None:
        """Build the graph the node will run."""

    @classmethod
    def _io_defining_function(cls) -> Callable:
        return cls.graph_creator

    _io_defining_function_uses_self = True

    @classmethod
    def _scrape_output_labels(cls):
        scraped_labels = super()._scrape_output_labels()

        if scraped_labels is not None:
            # Strip off the first argument, e.g. self.foo just becomes foo
            self_argument = list(cls._get_input_args().keys())[0]
            cleaned_labels = [
                re.sub(r"^" + re.escape(f"{self_argument}."), "", label)
                for label in scraped_labels
            ]
            if any("." in label for label in cleaned_labels):
                raise ValueError(
                    f"Tried to scrape cleaned labels for {cls.__name__}, but at least "
                    f"one of {cleaned_labels} still contains a '.' -- please provide "
                    f"explicit labels"
                )
            return cleaned_labels
        else:
            return scraped_labels

    def _prepopulate_ui_nodes_from_graph_creator_signature(self):
        ui_nodes = []
        for label, (type_hint, default) in self.preview_inputs().items():
            n = self.create.std.UserInput(
                default,
                label=label,
                parent=self,
            )
            n.inputs.user_input.type_hint = type_hint
            ui_nodes.append(n)
        return tuple(ui_nodes)

    def _purge_single_use_ui_nodes(self, ui_nodes):
        """
        We (may) create UI nodes based on the :meth:`graph_creator` signature;
        If these are connected to only a single node actually defined in the creator,
        they are superfluous, and we can remove them -- linking the macro input
        directly to the child node input.
        """
        remaining_ui_nodes = list(ui_nodes)
        for macro_input in self.inputs:
            target_node = macro_input.value_receiver.owner
            if (
                target_node in ui_nodes  # Value link is a UI node
                and target_node.channel.value_receiver is None  # That doesn't forward
                # its value directly to the output
                and len(target_node.channel.connections) <= 1  # And isn't forked to
                # multiple children
            ):
                if len(target_node.channel.connections) == 1:
                    macro_input.value_receiver = target_node.channel.connections[0]
                self.remove_child(target_node)
                remaining_ui_nodes.remove(target_node)
        return tuple(remaining_ui_nodes)

    @property
    def inputs(self) -> Inputs:
        return self._inputs

    @property
    def outputs(self) -> OutputsWithInjection:
        return self._outputs

    def _parse_remotely_executed_self(self, other_self):
        local_connection_data = [
            [(c, c.label, c.connections) for c in io_panel]
            for io_panel in [
                self.inputs,
                self.outputs,
                self.signals.input,
                self.signals.output,
            ]
        ]
        super()._parse_remotely_executed_self(other_self)

        for old_data, io_panel in zip(
            local_connection_data,
            [self.inputs, self.outputs, self.signals.input, self.signals.output],
            strict=False,
            # Get fresh copies of the IO panels post-update
        ):
            for original_channel, label, connections in old_data:
                new_channel = io_panel[label]  # Fetch it from the fresh IO panel
                new_channel.connections = connections
                for other_channel in connections:
                    self._replace_connection(
                        other_channel, original_channel, new_channel
                    )

    @staticmethod
    def _replace_connection(
        channel: Channel, old_connection: Channel, new_connection: Channel
    ):
        """Brute-force replace an old connection in a channel with a new one"""
        channel.connections = [
            c if c is not old_connection else new_connection for c in channel
        ]

    def _configure_graph_execution(self, ui_nodes):
        run_signals = self.disconnect_run()

        has_signals = len(run_signals) > 0
        has_starters = len(self.starting_nodes) > 0

        if has_signals and has_starters:
            # Assume the user knows what they're doing
            self._reconnect_run(run_signals)
            # Then put the UI upstream of the original starting nodes
            for n in self.starting_nodes:
                n << ui_nodes
            self.starting_nodes = ui_nodes if len(ui_nodes) > 0 else self.starting_nodes
        elif not has_signals and not has_starters:
            # Automate construction of the execution graph
            self.set_run_signals_to_dag_execution()
        else:
            raise ValueError(
                f"The macro {self.full_label} has {len(run_signals)} run signals "
                f"internally and {len(self.starting_nodes)} starting nodes. Either "
                f"the entire execution graph must be specified manually, or both run "
                f"signals and starting nodes must be left entirely unspecified for "
                f"automatic construction of the execution graph."
            )

    def _reconnect_run(self, run_signal_pairs_to_restore):
        self.disconnect_run()
        for pairs in run_signal_pairs_to_restore:
            pairs[0].connect(pairs[1])

    @property
    def _input_value_links(self):
        """
        Value connections between child output and macro in string representation based
        on labels.

        The string representation helps storage, and having it as a property ensures
        the name is protected.
        """
        return [
            (c.label, (c.value_receiver.owner.label, c.value_receiver.label))
            for c in self.inputs
        ]

    @property
    def _output_value_links(self):
        """
        Value connections between macro and child input in string representation based
        on labels.

        The string representation helps storage, and having it as a property ensures
        the name is protected.
        """
        return [
            ((c.owner.label, c.label), c.value_receiver.label)
            for child in self
            for c in child.outputs
            if c.value_receiver is not None
        ]

    def __getstate__(self):
        state = super().__getstate__()
        state["_input_value_links"] = self._input_value_links
        state["_output_value_links"] = self._output_value_links
        return state

    def __setstate__(self, state):
        # Purge value links from the state
        input_links = state.pop("_input_value_links")
        output_links = state.pop("_output_value_links")

        super().__setstate__(state)

        # Re-forge value links
        for inp, (child, child_inp) in input_links:
            self.inputs[inp].value_receiver = self.children[child].inputs[child_inp]

        for (child, child_out), out in output_links:
            self.children[child].outputs[child_out].value_receiver = self.outputs[out]

    @classmethod
    def _extra_info(cls) -> str:
        return getsource(cls.graph_creator)


@classfactory
def macro_node_factory(
    graph_creator: Callable,
    validate_output_labels: bool,
    use_cache: bool = True,
    /,
    *output_labels: str,
) -> type[Macro]:
    """
    Create a new :class:`Macro` subclass using the given graph creator function.

    Args:
        graph_creator (callable): Function to create the graph for this subclass of
            :class:`Macro`.
        validate_output_labels (bool): Whether to validate the output labels against
            the return values of the wrapped function.
        use_cache (bool): Whether nodes of this type should default to caching their
            values.
        output_labels (tuple[str, ...]): Optional labels for the :class:`Macro`'s
            outputs.

    Returns:
        type[Macro]: A new :class:`Macro` subclass.
    """
    return (  # type: ignore[return-value]
        graph_creator.__name__,
        (Macro,),  # Define parentage
        {
            "graph_creator": graph_creator,
            "__module__": graph_creator.__module__,
            "__qualname__": graph_creator.__qualname__,
            "_output_labels": None if len(output_labels) == 0 else output_labels,
            "_validate_output_labels": validate_output_labels,
            "__doc__": Macro._io_defining_documentation(graph_creator, "graph_creator"),
            "use_cache": use_cache,
        },
        {},
    )


@dispatch_output_labels
def as_macro_node(
    *output_labels: str, validate_output_labels: bool = True, use_cache: bool = True
):
    """
    Decorator to convert a function into a :class:`Macro` node.

    Args:
        *output_labels (str): Optional labels for the :class:`Macro`'s outputs.
        validate_output_labels (bool): Whether to validate the output labels.
        use_cache (bool): Whether nodes of this type should default to caching their
            values. (Default is True.)

    Returns:
        callable: A decorator that converts a function into a Macro node.
    """

    def decorator(graph_creator):
        macro_node_factory.clear(graph_creator.__name__)  # Force a fresh class
        factory_made = macro_node_factory(
            graph_creator, validate_output_labels, use_cache, *output_labels
        )
        factory_made._reduce_imports_as = (
            graph_creator.__module__,
            graph_creator.__qualname__,
        )
        factory_made.preview_io()
        return factory_made

    return decorator


def macro_node(
    graph_creator: Callable[..., GraphOutput | tuple[GraphOutput, ...] | None],
    *node_args,
    output_labels: str | tuple[str, ...] | None = None,
    validate_output_labels: bool = True,
    use_cache: bool = True,
    **node_kwargs,
):
    """
    Create and return a :class:`Macro` node instance using the given node function.

    Args:
        graph_creator (callable): Function to create the graph for the :class:`Macro`.
        node_args: Positional arguments for the :class:`Macro` initialization --
            parsed as node input data.
        output_labels (str | tuple[str, ...] | None): Labels for the :class:`Macro`'s
            outputs. Default is None, which scrapes these from the return statement in
            the decorated function's source code.
        validate_output_labels (bool): Whether to validate the output labels. Defaults
            to True.
        use_cache (bool): Whether this node should default to caching its values.
            (Default is True.)
        node_kwargs: Keyword arguments for the :class:`Macro` initialization --
            parsed as node input data when the keyword matches an input channel.

    Returns:
        Macro: An instance of the :class:`Macro` subclass.
    """
    if output_labels is None:
        output_labels = ()
    elif isinstance(output_labels, str):
        output_labels = (output_labels,)
    macro_node_factory.clear(graph_creator.__name__)  # Force a fresh class
    factory_made = macro_node_factory(
        graph_creator, validate_output_labels, use_cache, *output_labels
    )
    factory_made.preview_io()
    return factory_made(*node_args, **node_kwargs)



"""
Container classes for giving access to various workflow objects and tools
"""

from __future__ import annotations

from pyiron_snippets.dotdict import DotDict

from pyiron_workflow.executors import CloudpickleProcessPoolExecutor
from pyiron_workflow.nodes.function import as_function_node, function_node


class Creator:
    """
    A container class for providing access to various workflow objects.
    Gives access to various workflow tools.

    In addition to node objects, the creator also provides workflow-compliant executors
    for parallel processing.
    """

    def __init__(self):
        # Local cloudpickler
        self.CloudpickleProcessPoolExecutor = CloudpickleProcessPoolExecutor

        self.function_node = function_node

    @property
    def std(self):
        from pyiron_workflow.nodes import standard

        return standard

    @property
    def for_node(self):
        from pyiron_workflow.nodes.for_loop import for_node

        return for_node

    @property
    def while_node(self):
        from pyiron_workflow.nodes.while_loop import while_node

        return while_node

    @property
    def macro_node(self):
        from pyiron_workflow.nodes.macro import macro_node

        return macro_node

    @property
    def Workflow(self):
        from pyiron_workflow.workflow import Workflow

        return Workflow

    @property
    def meta(self):
        from pyiron_workflow.nodes.transform import inputs_to_list, list_to_outputs

        return DotDict(
            {
                inputs_to_list.__name__: inputs_to_list,
                list_to_outputs.__name__: list_to_outputs,
            }
        )

    @property
    def transformer(self):
        from pyiron_workflow.nodes.transform import (
            dataclass_node,
            inputs_to_dataframe,
            inputs_to_dict,
            inputs_to_list,
            list_to_outputs,
        )

        return DotDict(
            {
                f.__name__: f
                for f in [
                    dataclass_node,
                    inputs_to_dataframe,
                    inputs_to_dict,
                    inputs_to_list,
                    list_to_outputs,
                ]
            }
        )


class Wrappers:
    """
    A container class giving access to the decorators that transform functions to nodes.
    """

    as_function_node = staticmethod(as_function_node)

    @property
    def as_macro_node(self):
        from pyiron_workflow.nodes.macro import as_macro_node

        return as_macro_node

    @property
    def as_dataclass_node(self):
        from pyiron_workflow.nodes.transform import as_dataclass_node

        return as_dataclass_node


class HasCreator:
    """
    A mixin class for creator (including both class-like and decorator).
    """

    create = Creator()
    wrap = Wrappers()



"""
A bit of abstraction connecting generic storage routines to nodes.
"""

from __future__ import annotations

import pickle
from abc import ABC, abstractmethod
from collections.abc import Generator
from pathlib import Path
from typing import TYPE_CHECKING, Literal, TypeAlias

import bagofholding
import cloudpickle

if TYPE_CHECKING:
    from pyiron_workflow.node import Node


class TypeNotFoundError(ImportError):
    """
    Raised when you try to save a node, but importing its module and class give
    something other than its type.
    """


class FileTypeError(TypeError):
    """
    Raised when you're looking for a save file, and you find something, but the file
    appears to be the wrong type.
    """


def delete_files_and_directories_recursively(path: Path):
    """
    Recursively delete all files and directories in the given path.

    Args:
        path (Path): The path to the directory to delete.
    """
    if not path.exists():
        return
    for item in path.rglob("*"):
        if item.is_file():
            item.unlink()
        else:
            delete_files_and_directories_recursively(item)
    path.rmdir()


class StorageInterface(ABC):
    """
    Abstract base class defining the interface for saving, loading, and managing node
    storage.

    Public methods accept `**kwargs` so that specific implementations can have extra
    behaviour. In general, whatever can be set for the kwargs should also be able to
    be specified at initialization as a default, so that users can configure a storage
    back-end the way they want and then run it from there -- kwargs are just exposed to
    make special instances of use easier.
    """

    @abstractmethod
    def _save(self, node: Node, filename: Path, /, *args, **kwargs):
        """
        Save a node to file.

        Args:
            node (Node): The node to save
            filename (Path | None): The path to the save location (WITHOUT file
                extension.)
            **kwargs: Additional keyword arguments.
        """

    @abstractmethod
    def _load(self, filename: Path, /, *args, **kwargs) -> Node:
        """
        Instantiate a node from file.

        Args:
            filename (Path): The path to the file to load (WITHOUT file extension).
            **kwargs: Additional keyword arguments.

        Returns:
            Node: The node stored there.
        """

    @abstractmethod
    def _has_saved_content(self, filename: Path, /, *args, **kwargs) -> bool:
        """
        Check for a save file matching this storage interface.

        Args:
            filename (Path): The path to the file to look for (WITHOUT file extension).
            **kwargs: Additional keyword arguments.

        Returns:
            bool: Whether a commensurate file was found.
        """

    @abstractmethod
    def _delete(self, filename: Path, /, *args, **kwargs):
        """
        Remove an existing save-file for this backend.

        Args:
            filename (Path): The path to the file to delete (WITHOUT file extension).
            **kwargs: Additional keyword arguments.
        """

    def save(self, node: Node, filename: str | Path | None = None, **kwargs):
        """
        Save a node to file.

        Args:
            node (Node): The node to save
            filename (Path | None): The path to the save location (WITHOUT file
                extension.)
            **kwargs: Additional keyword arguments.
        """
        filename = self._parse_filename(
            node=node if filename is None else None,
            filename=filename,
        )
        filename.parent.mkdir(parents=True, exist_ok=True)

        try:
            self._save(node, filename, **kwargs)
        except Exception as e:
            raise e
        finally:
            # If nothing got written due to the exception, clean up the directory
            # (as long as there's nothing else in it)
            if not any(filename.parent.iterdir()):
                filename.parent.rmdir()

    def load(
        self, node: Node | None = None, filename: str | Path | None = None, **kwargs
    ) -> Node:
        """
        Load a node from a file.

        Args:
            node (Node | None): The node to load. Optional if filename is provided.
            filename (str | Path | None): The path to the file to load (without file
                extension). Uses the canonical filename based on the node's lexical
                path instead if this is None.
            **kwargs: Additional keyword arguments.

        Returns:
            Node: The loaded node.
        """
        return self._load(self._parse_filename(node=node, filename=filename), **kwargs)

    def has_saved_content(
        self,
        node: Node | None = None,
        filename: str | Path | None = None,
        **kwargs,
    ) -> bool:
        """
        Check if a file has contents related to a node.

        Args:
            node (Node | None): The node to check. Optional if filename is provided.
            filename (str | Path | None): The path to the file to check (without file
                extension). Optional if the node is provided.
            **kwargs: Additional keyword arguments.

        Returns:
            bool: True if contents exist, False otherwise.
        """
        return self._has_saved_content(
            self._parse_filename(node=node, filename=filename), **kwargs
        )

    def delete(
        self,
        node: Node | None = None,
        filename: str | Path | None = None,
        delete_even_if_not_empty: bool = False,
        **kwargs,
    ):
        """
        Delete a file associated with a node.

        Args:
            node (Node | None): The node whose associated file is to be deleted.
                Optional if filename is provided.
            filename (str | Path | None): The path to the file to delete (without file
                extension). Optional if the node is provided.
            **kwargs: Additional keyword arguments.
        """
        filename = self._parse_filename(node=node, filename=filename)
        if self._has_saved_content(filename, **kwargs):
            self._delete(filename, **kwargs)
        if delete_even_if_not_empty:
            delete_files_and_directories_recursively(filename.parent)
        elif filename.parent.exists() and not any(filename.parent.iterdir()):
            filename.parent.rmdir()

    def _parse_filename(
        self, node: Node | None, filename: str | Path | None = None
    ) -> Path:
        """
        Make sure the node xor filename was provided, and if it's the node, convert it
        into a canonical filename by exploiting the node's lexical path.
        """
        if node is None and filename is None:
            raise ValueError(
                "At least one of node or filename must be specified, or we can't know "
                "where to load from."
            )
        elif node is None and filename is not None:
            if isinstance(filename, Path):
                return filename
            elif isinstance(filename, str):
                return Path(filename)
            else:
                raise TypeError(
                    f"Expected filename to be str, pathlib.Path, or None, but got "
                    f"{filename}"
                )
        elif node is not None and filename is None:
            return node.as_path() / self.__class__.__name__.lower()
        elif node is not None and filename is not None:
            raise ValueError(
                f"Both the node ({node.full_label}) and filename ({filename}) were "
                f"specified for loading -- please only specify one or the other."
            )
        else:
            raise AssertionError(
                "This is an unreachable state -- we have covered all four cases of the "
                "boolean `is (not) None` square."
            )


class PickleStorage(StorageInterface):
    _PICKLE = ".pckl"
    _CLOUDPICKLE = ".cpckl"

    def __init__(self, cloudpickle_fallback: bool = True):
        self.cloudpickle_fallback = cloudpickle_fallback

    def _fallback(self, cpf: bool | None) -> bool:
        return self.cloudpickle_fallback if cpf is None else cpf

    def _save(
        self, node: Node, filename: Path, /, cloudpickle_fallback: bool | None = None
    ):
        if not self._fallback(cloudpickle_fallback) and not node.import_ready:
            raise TypeNotFoundError(
                f"{node.label} cannot be saved with the storage interface "
                f"{self.__class__.__name__} because it (or one of its children) has "
                f"a type that cannot be imported. Is this node defined inside <locals>? "
                f"\n"
                f"Import readiness report: \n"
                f"{node.report_import_readiness()}"
            )

        attacks = [(self._PICKLE, pickle.dump)]
        if self._fallback(cloudpickle_fallback):
            attacks += [(self._CLOUDPICKLE, cloudpickle.dump)]

        e: Exception | None = None
        for suffix, save_method in attacks:
            e = None
            p = filename.with_suffix(suffix)
            try:
                with open(p, "wb") as filehandle:
                    save_method(node, filehandle)
                return
            except Exception as ee:
                e = ee
                p.unlink(missing_ok=True)
        if e is not None:
            raise e

    def _load(
        self, filename: Path, /, cloudpickle_fallback: bool | None = None
    ) -> Node:
        attacks = [(self._PICKLE, pickle.load)]
        if self._fallback(cloudpickle_fallback):
            attacks += [(self._CLOUDPICKLE, cloudpickle.load)]

        for suffix, load_method in attacks:
            p = filename.with_suffix(suffix)
            if p.is_file():
                with open(p, "rb") as filehandle:
                    inst = load_method(filehandle)
                return inst
        raise FileNotFoundError(f"Could not load {filename}, no such file found.")

    def _delete(self, filename: Path, /, cloudpickle_fallback: bool | None = None):
        suffixes = (
            [self._PICKLE, self._CLOUDPICKLE]
            if self._fallback(cloudpickle_fallback)
            else [self._PICKLE]
        )
        for suffix in suffixes:
            filename.with_suffix(suffix).unlink(missing_ok=True)

    def _has_saved_content(
        self, filename: Path, /, cloudpickle_fallback: bool | None = None
    ) -> bool:
        suffixes = (
            [self._PICKLE, self._CLOUDPICKLE]
            if self._fallback(cloudpickle_fallback)
            else [self._PICKLE]
        )
        return any(filename.with_suffix(suffix).exists() for suffix in suffixes)


class H5BagStorage(StorageInterface):
    """
    Storage using :class:`bagofholding.H5Bag`.
    """

    @staticmethod
    def _ensure_file_extension(filename: Path | str) -> str:
        suffix = ".h5"
        fname = str(filename)
        return fname if fname.endswith(suffix) else fname + suffix

    def _save(self, node: Node, filename: Path, /, *args, **kwargs):
        if not node.import_ready:
            raise TypeNotFoundError(
                f"{node.label} cannot be saved with the storage interface "
                f"{self.__class__.__name__} because it (or one of its children) has "
                f"a type that cannot be imported. Is this node defined inside <locals>? "
                f"\n"
                f"Import readiness report: \n"
                f"{node.report_import_readiness()}"
            )
        bagofholding.H5Bag.save(
            node, self._ensure_file_extension(filename), *args, **kwargs
        )

    def _load(self, filename: Path, /, *args, **kwargs) -> Node:
        bag = bagofholding.H5Bag(self._ensure_file_extension(filename))
        return bag.load(*args, **kwargs)

    def _has_saved_content(self, filename: Path) -> bool:
        alleged_location = Path(self._ensure_file_extension(filename))
        if not alleged_location.exists():
            return False
        try:
            bagofholding.H5Bag(alleged_location)
        except Exception as e:
            raise FileTypeError(
                f"Found a file at {alleged_location}, but it could not be resolved as "
                f"a {bagofholding.H5Bag.__name__} bag."
            ) from e
        return True

    def _delete(self, filename: Path):
        Path(self._ensure_file_extension(filename)).unlink(missing_ok=True)


BackendIdentifier: TypeAlias = Literal["h5bag", "pickle"]

_standard_backends: dict[BackendIdentifier, type[StorageInterface]] = {
    "h5bag": H5BagStorage,
    "pickle": PickleStorage,
}


def available_backends(
    backend: BackendIdentifier | StorageInterface | None = None,
    only_requested: bool = False,
) -> Generator[StorageInterface, None, None]:
    """
    A generator for accessing available :class:`StorageInterface` instances, starting
    with the one requested.

    Args:
        backend (BackendIdentifier | StorageInterface | None): The interface to yield
            first.
        only_requested (bool): Stop after yielding whatever was specified by
            :param:`backend`.

    Yields:
        StorageInterface: An interface for serializing :class:`Node`.
    """
    backend_instance = (
        _standard_backends.get(backend, PickleStorage)()
        if isinstance(backend, str)
        else backend
    )

    if backend_instance is not None:
        yield backend_instance
        if only_requested:
            return

    yield from (v() for k, v in _standard_backends.items() if k != backend)



from __future__ import annotations

from abc import ABC

from pyiron_snippets.colors import SeabornColors

from pyiron_workflow.channels import InputData
from pyiron_workflow.io import Inputs
from pyiron_workflow.mixin.injection import (
    OutputDataWithInjection,
    OutputsWithInjection,
)
from pyiron_workflow.mixin.preview import HasIOPreview
from pyiron_workflow.node import Node


class StaticNode(Node, HasIOPreview, ABC):
    """
    A node whose IO specification is available at the class level.

    Actual IO is then constructed from the preview at instantiation.
    """

    def _setup_node(self) -> None:
        super()._setup_node()

        self._inputs = Inputs(
            *[
                InputData(
                    label=label,
                    owner=self,
                    default=default,
                    type_hint=type_hint,
                )
                for label, (type_hint, default) in self.preview_inputs().items()
            ]
        )

        self._outputs = OutputsWithInjection(
            *[
                OutputDataWithInjection(
                    label=label,
                    owner=self,
                    type_hint=hint,
                )
                for label, hint in self.preview_outputs().items()
            ]
        )

    @property
    def inputs(self) -> Inputs:
        return self._inputs

    @property
    def outputs(self) -> OutputsWithInjection:
        return self._outputs

    def display_state(self, state=None, ignore_private=True):
        state = dict(self.__getstate__()) if state is None else state
        self._make_entry_public(state, "_inputs", "inputs")
        self._make_entry_public(state, "_outputs", "outputs")
        self._make_entry_public(state, "_signals", "signals")
        return super().display_state(state=state, ignore_private=ignore_private)

    @property
    def color(self) -> str:
        """For drawing the graph"""
        return SeabornColors.pink



"""
Tools specifically for the test suite, not intended for general use.
"""

import sys
from pathlib import Path


def ensure_tests_in_python_path():
    """So that you can import from the static module"""
    path_to_tests = Path(__file__).parent.parent / "tests"
    as_string = str(path_to_tests.resolve())

    if as_string not in sys.path:
        sys.path.append(as_string)



from __future__ import annotations

import inspect
from abc import ABC, abstractmethod
from collections.abc import Callable
from typing import Any

from pyiron_snippets.colors import SeabornColors
from pyiron_snippets.factory import classfactory

from pyiron_workflow.mixin.preview import ScrapesIO
from pyiron_workflow.nodes.multiple_distpatch import dispatch_output_labels
from pyiron_workflow.nodes.static_io import StaticNode


class Function(StaticNode, ScrapesIO, ABC):
    """
    Function nodes wrap an arbitrary python function.

    Actual function node instances can either be instances of the base node class, in
    which case the callable node function *must* be provided OR they can be instances
    of children of this class which provide the node function as a class-level method.
    Those children may define some or all of the node behaviour at the class level, and
    modify their signature accordingly so this is not available for alteration by the
    user, e.g. the node function and output labels may be hard-wired.

    Although not strictly enforced, it is a best-practice that where possible, function
    nodes should be both functional (always returning the same output given the same
    input) and idempotent (not modifying input data in-place, but creating copies where
    necessary and returning new objects as output).
    Further, functions with multiple return branches that return different types or
    numbers of return values may or may not work smoothly, depending on the details.

    Promises:

    - IO channels are constructed automatically from the wrapped function
        - This includes type hints (if any)
        - This includes defaults (if any)
        - By default one channel is created for each returned value (from a tuple)...
        - Output channel labels are taken from the returned value, but may be overriden
        - A single tuple output channel can be forced by manually providing exactly one
            output label
    - Running the node executes the wrapped function and returns its result
    - Input updates can be made with `*args` as well as the usual `**kwargs`, following
        the same input order as the wrapped function.
    - A default label can be scraped from the name of the wrapped function

    Examples:
        At the most basic level, to use nodes all we need to do is provide the
        `Function` class with a function and labels for its output, like so:

        >>> from pyiron_workflow import function_node
        >>>
        >>> def mwe(x, y):
        ...     return x+1, y-1
        >>>
        >>> plus_minus_1 = function_node(mwe)
        >>>
        >>> print(plus_minus_1.outputs["x+1"])
        NOT_DATA

        There is no output because we haven't given our function any input, it has
        no defaults, and we never ran it! So outputs have the channel default value of
        `NOT_DATA` -- a special non-data singleton (since `None` is sometimes a
        meaningful value in python).

        We'll run into a hiccup if we try to set only one of the inputs and force the
        run:

        >>> plus_minus_1.inputs.x = 2
        >>> try:
        ...     plus_minus_1.run()
        ... except ValueError as e:
        ...     print("ValueError:", e.args[0])
        ValueError: mwe received a run command but is not ready. The node should be neither running nor failed, and all input values should conform to type hints.
        mwe readiness report:
        ready: False
        running: False
        failed: False
        inputs.x: True
        inputs.y: False
        <BLANKLINE>

        We are able to check this without trying and failing by looking at the
        readiness report:

        >>> print(plus_minus_1.readiness_report)
        mwe readiness report:
        ready: False
        running: False
        failed: False
        inputs.x: True
        inputs.y: False
        <BLANKLINE>

        This is because the second input (`y`) still has no input value -- indicated in
        the error message -- so we can't do the sum between `NOT_DATA` and `2`.

        Once we update `y`, all the input is ready we will be allowed to proceed to a
        `run()` call, which succeeds and updates the output.
        The final thing we need to do is disable the `failed` status we got from our
        last run call

        >>> plus_minus_1.failed = False
        >>> plus_minus_1.inputs.y = 3
        >>> out = plus_minus_1.run()
        >>> plus_minus_1.outputs.to_value_dict()
        {'x+1': 3, 'y-1': 2}

        We can also, optionally, provide initial values for some or all of the input
        and labels for the output:

        >>> plus_minus_1 = function_node(mwe, output_labels=("p1", "m1"),  x=1)
        >>> plus_minus_1.inputs.y = 2
        >>> out = plus_minus_1.run()
        >>> out
        (2, 1)

        Input data can be provided to both initialization and on call as ordered args
        or keyword kwargs.
        When running the node (or any alias to run like pull, execute, or just calling
        the node), the output of the wrapped function is returned:

        >>> plus_minus_1(2, y=3)
        (3, 2)

        We can make our node even more sensible by adding type
        hints (and, optionally, default values) when defining the function that the
        node wraps.
        The node will automatically figure out defaults and type hints for the IO
        channels from inspection of the wrapped function.

        In this example, note the mixture of old-school (`typing.Union`) and new (`|`)
        type hints as well as nested hinting with a union-type inside the tuple for the
        return hint.
        Our treatment of type hints is **not infinitely robust**, but covers a wide
        variety of common use cases.
        Note that getting "good" (i.e. dot-accessible) output labels can be achieved by
        using good variable names and returning those variables instead of using
        :param:`output_labels`.
        If we try to assign a value of the wrong type, it will raise an error:

        >>> from typing import Union
        >>>
        >>> def hinted_example(
        ...     x: Union[int, float],
        ...     y: int | float = 1
        ... ) -> tuple[int, int | float]:
        ...     p1, m1 = x+1, y-1
        ...     return p1, m1
        >>>
        >>> plus_minus_1 = function_node(hinted_example)
        >>> try:
        ...     plus_minus_1.inputs.x =  "not an int or float"
        ... except TypeError as e:
        ...     print("TypeError:", e.args[0])
        TypeError: The channel /hinted_example.x cannot take the value `not an int or float` (<class 'str'>) because it is not compliant with the type hint typing.Union[int, float]

        We can turn off type hinting with the `strict_hints` boolean property, or just
        circumvent the type hinting by applying the new data directly to the private
        `_value` property.
        In the latter case, we'd still get a readiness error when we try to run and
        the ready check sees that the data doesn't conform to the type hint:

        >>> plus_minus_1.inputs.x._value =  "not an int or float"
        >>> try:
        ...     plus_minus_1.run()
        ... except ValueError as e:
        ...     print("ValueError:", e.args[0])
        ValueError: hinted_example received a run command but is not ready. The node should be neither running nor failed, and all input values should conform to type hints.
        hinted_example readiness report:
        ready: False
        running: False
        failed: False
        inputs.x: False
        inputs.y: True
        <BLANKLINE>

        Here, even though all the input has data, the node sees that some of it is the
        wrong type and so (by default) the run raises an error right away.
        This causes the failure to come earlier because we stop the node from running
        and throwing an error because it sees that the channel (and thus node) is not
        ready:

        >>> plus_minus_1.ready, plus_minus_1.inputs.x.ready, plus_minus_1.inputs.y.ready
        (False, False, True)

        In these examples, we've instantiated nodes directly from the base
        :class:`Function` class, and populated their input directly with data.
        In practice, these nodes are meant to be part of complex workflows; that means
        both that you are likely to have particular nodes that get heavily re-used, and
        that you need the nodes to pass data to each other.

        For reusable nodes, we want to create a sub-class of :class:`Function`
        that fixes some of the node behaviour -- i.e. the :meth:`node_function`.

        This can be done most easily with the :func:`as_function_node` decorator, which
        takes a function and returns a node class. This can be used in the usual way,
        but the decorator itself also optionally accepts some arguments. Namely, it
        also allows us to provide labels for the return values, :param:output_labels,
        which are otherwise scraped from the text of the function definition:

        >>> from pyiron_workflow import as_function_node
        >>>
        >>> @as_function_node("p1", "m1")
        ... def my_mwe_node(
        ...     x: int | float, y: int | float = 1
        ... ) -> tuple[int | float, int | float]:
        ...     return x+1, y-1
        >>>
        >>> node_instance = my_mwe_node(x=0)
        >>> node_instance(y=0)
        (1, -1)

        Where we've passed the output labels and class arguments to the decorator,
        and inital values to the newly-created node class (`my_mwe_node`) at
        instantiation.
        Because we provided a good initial value for `x`, we get our result right away.

        Using the decorator is the recommended way to create new node classes, but this
        magic is just equivalent to creating a child class with the `node_function`
        already defined as a `staticmethod`:

        >>> from typing import Literal, Optional
        >>> from pyiron_workflow.api import Function
        >>>
        >>> class AlphabetModThree(Function):
        ...
        ...     @staticmethod
        ...     def node_function(i: int) -> Literal["a", "b", "c"]:
        ...         letter = ["a", "b", "c"][i % 3]
        ...         return letter


        Finally, let's put it all together by using both of these nodes at once.
        Instead of setting input to a particular data value, we'll set it to
        be another node's output channel, thus forming a connection.
        At the end of the day, the graph will also need to know about the execution
        flow, but in most cases (directed acyclic graphs -- DAGs), this can be worked
        out automatically by the topology of data connections.
        Let's put together a couple of nodes and then run in a "pull" paradigm to get
        the final node to run everything "upstream" then run itself:

        >>> @as_function_node
        ... def adder_node(x: int = 0, y: int = 0) -> int:
        ...     sum = x + y
        ...     return sum
        >>>
        >>> adder = adder_node(x=1)
        >>> alpha = AlphabetModThree(i=adder.outputs.sum)
        >>> print(alpha())
        b
        >>> adder.inputs.y = 1
        >>> print(alpha())
        c
        >>> adder.inputs.x = 0
        >>> adder.inputs.y = 0
        >>> print(alpha())
        a

        Alternatively, execution flows can be specified manualy by connecting
        `.signals.input.run` and `.signals.output.ran` channels, either by their
        `.connect` method or by assignment (both cases just like data chanels), or by
        some syntactic sugar using the `>` operator.
        Then we can use a "push" paradigm with the `run` command to force execution
        forwards through the graph to get an end result.
        This is a bit more verbose, but a necessary tool for more complex situations
        (like cyclic graphs).
        Here's our simple example from above using this other paradigm:

        >>> @as_function_node
        ... def adder_node(x: int = 0, y: int = 0) -> int:
        ...     sum = x + y
        ...     return sum
        >>>
        >>> adder = adder_node()
        >>> alpha = AlphabetModThree(i=adder.outputs.sum)
        >>> _ = adder >> alpha
        >>> # We catch and ignore output -- it's needed for chaining, but screws up
        >>> # doctests -- you don't normally need to catch it like this!
        >>> out = adder.run(x=1)
        >>> print(alpha.outputs.letter)
        b
        >>> out = adder.run(y=1)
        >>> print(alpha.outputs.letter)
        c
        >>> adder.inputs.x = 0
        >>> adder.inputs.y = 0
        >>> out = adder.run()
        >>> print(alpha.outputs.letter)
        a

        To see more details on how to use many nodes together, look at the
        :class:`Workflow` class.

    Comments:
        Using the `self` argument for function nodes is not fully supported; it will
        raise an error when combined with an executor, and otherwise behaviour is not
        guaranteed.
    """

    @staticmethod
    @abstractmethod
    def node_function(**kwargs) -> Callable:
        """What the node _does_."""

    @classmethod
    def _io_defining_function(cls) -> Callable:
        return cls.node_function

    @classmethod
    def _build_outputs_preview(cls) -> dict[str, Any]:
        preview = super()._build_outputs_preview()
        return preview if len(preview) > 0 else {"None": type(None)}
        # If clause facilitates functions with no return value

    def _on_run(self, **kwargs):
        return self.node_function(**kwargs)

    @property
    def run_args(self) -> tuple[tuple, dict]:
        kwargs = self.inputs.to_value_dict()
        return (), kwargs

    def process_run_result(self, function_output: Any | tuple) -> Any | tuple:
        """
        Take the results of the node function, and use them to update the node output.
        """
        for out, value in zip(
            self.outputs,
            (function_output,) if len(self.outputs) == 1 else function_output,
            strict=False,
        ):
            out.value = value
        return self._outputs_to_run_return()

    def _outputs_to_run_return(self):
        output = tuple(self.outputs.to_value_dict().values())
        if len(output) == 1:
            output = output[0]
        return output

    @property
    def color(self) -> str:
        """For drawing the graph"""
        return SeabornColors.green

    @classmethod
    def _extra_info(cls) -> str:
        return inspect.getsource(cls.node_function)


@classfactory
def function_node_factory(
    node_class_qualname: str,
    node_class_module_name: str,
    node_function: Callable,
    validate_output_labels: bool,
    use_cache: bool = True,
    /,
    *output_labels,
) -> type[Function]:
    """
    Create a new :class:`Function` node class based on the given node function. This
    function gets executed on each :meth:`run` of the resulting function.

    Args:
        node_function (callable): The function to be wrapped by the node.
        validate_output_labels (bool): Flag to indicate if output labels should be
            validated.
        use_cache (bool): Whether nodes of this type should default to caching their
            values.
        *output_labels: Optional labels for the function's output channels.

    Returns:
        type[Node]: A new node class.
    """
    node_class_name = node_class_qualname.rsplit(".", 1)[-1]
    return (  # type: ignore[return-value]
        node_class_name,
        (Function,),  # Define parentage
        {
            "node_function": staticmethod(node_function),
            "__module__": node_class_module_name,
            "__qualname__": node_class_qualname,
            "_output_labels": None if len(output_labels) == 0 else output_labels,
            "_validate_output_labels": validate_output_labels,
            "__doc__": Function._io_defining_documentation(
                node_function, "node_function"
            ),
            "use_cache": use_cache,
        },
        {},
    )


@dispatch_output_labels
def as_function_node(
    *output_labels: str,
    validate_output_labels=True,
    use_cache=True,
):
    """
    Decorator to create a new :class:`Function` node class from a given function. This
    function gets executed on each :meth:`run` of the resulting function.

    Args:
        *output_labels (str): Optional labels for the function's output channels.
        validate_output_labels (bool): Flag to indicate if output labels should be
            validated against the return values in the function node source code.
            Defaults to True.
        use_cache (bool): Whether nodes of this type should default to caching their
            values. (Default is True.)

    Returns:
        Callable: A decorator that converts a function into a :class:`Function` node
            subclass.
    """

    def decorator(node_function) -> type[Function]:
        function_node_factory.clear(node_function.__name__)  # Force a fresh class
        factory_made = function_node_factory(
            node_function.__qualname__,
            node_function.__module__,
            node_function,
            validate_output_labels,
            use_cache,
            *output_labels,
        )
        factory_made._reduce_imports_as = (
            node_function.__module__,
            node_function.__qualname__,
        )
        factory_made.preview_io()
        return factory_made

    return decorator


def to_function_node(
    node_class_name,
    node_function,
    *output_labels,
    validate_output_labels: bool = True,
    use_cache: bool = True,
    scope: dict[str, type] | None = None,
) -> type[Function]:
    """
    Create a new :class:`Function` node class from an existing function.
    Useful when the function does not exist in a context where you are free to
    decorate it, e.g.

    >>> import numpy as np
    >>> from pyiron_workflow.nodes.function import to_function_node
    >>>
    >>> Trapz = to_function_node("Trapz", np.trapz, "trapz")
    >>> Trapz.preview_io()
    {'inputs': {'y': (None, NOT_DATA), 'x': (None, None), 'dx': (None, 1.0), 'axis': (None, -1)}, 'outputs': {'trapz': None}}

    We still have two requirements on functions converted in this way:
    - The function must be inspectable
        - e.g. :func:`numpy.arange` fails this requirement
    - The function must not use protected or argument names (as with decorated
        functions)
        - e.g. variadics `*args` and `**kwargs`
    - The function must have a single return value (as with decorated functions)

    Otherwise you will need to explicitly write a decorated function that wraps your
    desired function.

    Because nodes convert type hints to actual python objects for strict type checking,
    we also need to provide non-builting type hints in the scope of the new node class
    (for the benefit of an underlying `inspect.signature(..., eval_str=True)` call).
    E.g., this function hints that it returns `set[Node]`, so while the new class is
    being created it will need to know how to parse the `"Node"` string type hint
    into an object. We do this by providing the `Node` class in its `scope` dictionary
    (it already knows what a `set` is because this is just a python built-in type):

    >>> from pyiron_workflow.topology import get_nodes_in_data_tree
    >>> from pyiron_workflow.node import Node
    >>>
    >>> GetNodesInDataTree = to_function_node(
    ...     "GetNodesInDataTree",
    ...     get_nodes_in_data_tree,
    ...     "nodes_set",  # Just a nice label for the output
    ...     scope={"Node": Node},
    ... )
    >>>
    >>> print(GetNodesInDataTree.preview_io())
    {'inputs': {'node': (<class 'pyiron_workflow.node.Node'>, NOT_DATA)}, 'outputs': {'nodes_set': set[pyiron_workflow.node.Node]}}


    Args:
        node_class_name (str): The name of the new class -- MUST be manually matched to
            the variable name to which the class is being assigned, or the class won't
            be importable.
        node_function (Callable): The function to be wrapped by the node.
        *output_labels (str): Optional labels for the function's output channels.
        validate_output_labels (bool): Flag to indicate if output labels should be
            validated against the return values in the function node source code.
            Defaults to True.
        use_cache (bool): Whether nodes of this type should default to caching their
            values. (Default is True.)

    Returns:
        type[Function]: A new node class subclassing :class:`Function`.
    """
    # Inspect the caller's frame in order to extract the module where this is being used
    frame = inspect.stack()[1]
    module = inspect.getmodule(frame[0])
    node_class_module_name = module.__name__ if module else None

    function_node_factory.clear(node_class_name)  # Force a fresh class
    factory_made = function_node_factory(
        node_class_name,
        node_class_module_name,
        node_function,
        validate_output_labels,
        use_cache,
        *output_labels,
    )
    factory_made._extra_type_hint_scope = scope
    factory_made.preview_io()
    return factory_made


def function_node(
    node_function: Callable,
    *node_args,
    output_labels: str | tuple[str, ...] | None = None,
    validate_output_labels: bool = True,
    use_cache: bool = True,
    **node_kwargs,
) -> Function:
    """
    Create and initialize a new instance of a :class:`Function` node.

    Args:
        node_function (callable): The function to be wrapped by the node.
        *node_args: Positional arguments for the :class:`Function` initialization --
            parsed as node input data.
        output_labels (str | tuple | Noen): Labels for the function's output
            channels. Defaults to None, which tries to parse these from the return
            statement.
        validate_output_labels (bool): Flag to indicate if output labels should be
            validated against the return values in the function source code. Defaults
            to True. Disabling this may be useful if the source code is not available
            or if the function has multiple return statements.
        use_cache (bool): Whether this node should default to caching its values.
            (Default is True.)
        **node_kwargs: Keyword arguments for the :class:`Function` initialization --
            parsed as node input data when the keyword matches an input channel.

    Returns:
        Function: An instance of the generated :class:`Function` node subclass.
    """
    if output_labels is None:
        output_labels = ()
    elif isinstance(output_labels, str):
        output_labels = (output_labels,)
    function_node_factory.clear(node_function.__name__)  # Force a fresh class
    factory_made = function_node_factory(
        node_function.__qualname__,
        node_function.__module__,
        node_function,
        validate_output_labels,
        use_cache,
        *output_labels,
    )
    factory_made.preview_io()
    return factory_made(*node_args, **node_kwargs)
